#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import pandas as pd
from qtpy.QtCore import QFileInfo
from .parserdata import ParserData
from .parser import Parser


class CSVParser(Parser):
    @staticmethod
    def load_file(path: str, *args, **kwargs) -> ParserData:
        assert path.lower().endswith('csv')

        # Metadata
        properties = {
            'name': QFileInfo(path).completeBaseName(),
            'extension': QFileInfo(path).suffix()
        }

        with open(path, 'r') as f:
            data = ParserData()
            data.data = pd.read_csv(f)
            data.properties = properties

            return data

    @staticmethod
    def save_file(*args, **kwargs) -> None:
        path = kwargs.get('path')

        if path is None:
            raise KeyError('Path missing.')

        data = pd.DataFrame(kwargs.get('data', {}))
        data.to_csv(path, index=False)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
import traceback

from OpenGL.GL import *

from qtpy.QtCore import QPoint
from qtpy.QtCore import Signal
from qtpy.QtCore import QFileInfo
from qtpy.QtGui import QMatrix4x4
from qtpy.QtGui import QPixmap
from qtpy.QtGui import QRegion
from qtpy.QtGui import QVector4D
from qtpy.QtWidgets import QOpenGLWidget

from .collections.glaxiscollection import GLAxisCollection
from .collections.glbackgroundcollection import GLBackgroundCollection
from .collections.gldrawablecollection import GLDrawableCollection

from .drawables.gldrawable import GLDrawable
from .drawables.axisgl import AxisGL
from .drawables.backgroundgl import BackgroundGL
from .drawables.blockgl import BlockGL
from .drawables.linegl import LineGL
from .drawables.meshgl import MeshGL
from .drawables.pointgl import PointGL
from .drawables.tubegl import TubeGL

from .fpscounter import FPSCounter

from ..controller.detectionmode import DetectionMode
from ..controller.normalmode import NormalMode
from ..controller.slicemode import SliceMode
from ..controller.measurementmode import MeasurementMode

from ..model import utils
from ..model.model import Model


class IntegrableViewer(QOpenGLWidget):
    # Signals
    signal_file_modified = Signal()
    signal_load_success = Signal(int)
    signal_load_failure = Signal()
    signal_export_success = Signal(int)
    signal_export_failure = Signal()

    signal_mesh_clicked = Signal(object)
    signal_mesh_distances = Signal(object)
    signal_mesh_sliced = Signal(object)
    signal_blocks_sliced = Signal(object)

    signal_mode_updated = Signal(str)
    signal_fps_updated = Signal(float)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)

        # Model
        self.model = Model()

        # Drawable elements
        self.axis_collection = GLAxisCollection(self)
        self.background_collection = GLBackgroundCollection(self)
        self.drawable_collection = GLDrawableCollection(self)

        self.axis = AxisGL(id='AXIS')
        self.axis.add_observer(self)
        self.axis_collection.add(self.axis)

        self.bg = BackgroundGL(id='BG')
        self.bg.add_observer(self)
        self.background_collection.add(self.bg)

        # Signals for viewer (self)
        self.signal_mode_updated.connect(lambda m: print(f'MODE: {m}'))
        self.signal_file_modified.connect(self.recreate)

        # Controllers
        self.current_mode = None
        self.controllers = {}

        self.add_controller(NormalMode, 'normal')
        self.add_controller(DetectionMode, 'detection')
        self.add_controller(SliceMode, 'slice')
        self.add_controller(MeasurementMode, 'measurement')
        self.set_normal_mode()

        # Camera/World/Projection
        self.camera = QMatrix4x4()
        self.world = QMatrix4x4()
        self.proj = QMatrix4x4()

        # FPS Counter
        self.fps_counter = FPSCounter()

        # Initial positions and rotations
        self.rotation_center = [0.0, 0.0, 0.0]
        self.rotation_angle = [0.0, 0.0, 0.0]
        self.camera_position = [0.0, 0.0, 200.0]

        # Extra information
        self._turbo_rendering = False
        self._autofit_to_screen = False

        self.fov = 45.0
        self.smoothness = 2.0  # Bigger => smoother (but slower) rotations
        self.projection_mode = 'perspective'  # 'perspective'/'orthographic'

    """
    Properties
    """
    @property
    def last_id(self) -> int:
        return self.drawable_collection.last_id

    @property
    def last_drawable(self) -> GLDrawable:
        return self.get_drawable(self.last_id)

    @property
    def off_center(self) -> np.ndarray:
        return self.camera_position - self.rotation_center

    @property
    def camera_position(self) -> np.ndarray:
        return np.array([self.xCameraPos, self.yCameraPos, self.zCameraPos])

    @property
    def rotation_angle(self) -> np.ndarray:
        return np.array([self.xCenterRot, self.yCenterRot, self.zCenterRot])

    @property
    def rotation_center(self) -> np.ndarray:
        return np.array([self.xCenterPos, self.yCenterPos, self.zCenterPos])

    @camera_position.setter
    def camera_position(self, pos: list) -> None:
        self.xCameraPos, self.yCameraPos, self.zCameraPos = pos

    @rotation_angle.setter
    def rotation_angle(self, rot: list) -> None:
        self.xCenterRot, self.yCenterRot, self.zCenterRot = rot

    @rotation_center.setter
    def rotation_center(self, center: list) -> None:
        self.xCenterPos, self.yCenterPos, self.zCenterPos = center

    """
    Turbo/Autofit
    """
    def get_turbo_status(self) -> bool:
        return self._turbo_rendering

    def get_autofit_status(self) -> bool:
        return self._autofit_to_screen

    def set_turbo_status(self, status: bool) -> None:
        if status:
            print('WARNING: Turbo-rendering might leave you without memory!')

        self._turbo_rendering = status
        self.update_turbo()

    def set_autofit_status(self, status: bool) -> None:
        self._autofit_to_screen = status
        self.update_autofit()

    def update_turbo(self) -> None:
        for d in self.get_all_drawables():
            d.is_boostable = self.get_turbo_status()

    def update_autofit(self) -> None:
        if self.get_autofit_status():
            self.fit_to_screen()

    """
    Projections
    """
    def perspective_projection(self) -> None:
        self.projection_mode = 'perspective'
        self.update()

    def orthographic_projection(self) -> None:
        self.projection_mode = 'orthographic'
        self.update()

    """
    Generator methods
    """
    def generate_drawable(self, generator):
        try:
            drawable = generator()
            self.signal_load_success.emit(drawable.id)

            return drawable
        except Exception:
            self.signal_load_failure.emit()
            traceback.print_exc()

            return None

    def generate_mesh(self, generator, *args, **kwargs) -> MeshGL:
        return self.generate_drawable(lambda: MeshGL(generator(*args, **kwargs), *args, **kwargs))

    def generate_blocks(self, generator, *args, **kwargs) -> BlockGL:
        return self.generate_drawable(lambda: BlockGL(generator(*args, **kwargs), *args, **kwargs))

    def generate_points(self, generator, *args, **kwargs) -> PointGL:
        return self.generate_drawable(lambda: PointGL(generator(*args, **kwargs), *args, **kwargs))

    def generate_lines(self, generator, *args, **kwargs) -> LineGL:
        return self.generate_drawable(lambda: LineGL(generator(*args, **kwargs), *args, **kwargs))

    def generate_tubes(self, generator, *args, **kwargs) -> TubeGL:
        return self.generate_drawable(lambda: TubeGL(generator(*args, **kwargs), *args, **kwargs))

    """
    Register methods
    """
    def register_drawable(self, drawable):
        if drawable is not None:
            drawable.add_observer(self)
            self.drawable_collection.add(drawable)
            self.signal_file_modified.emit()

        return drawable

    def register_folder(self, path: str, generator, *args, **kwargs) -> list:
        path_list = self.model.get_paths_from_directory(path)

        # We'll block viewer.register_drawable's signal_file_modified until last item has been loaded.
        self.blockSignals(True)
        loaded = [d for d in [generator(path, *args, **kwargs) for path in path_list] if d is not None]
        self.blockSignals(False)

        # We'll manually emit the signals we didn't emit before.
        if len(loaded) > 0:
            self.signal_file_modified.emit()
            self.signal_load_success.emit(self.last_id)
        else:
            self.signal_load_failure.emit()

        return loaded

    """
    Load methods by arguments
    """
    def mesh(self, *args, **kwargs) -> MeshGL:
        return self.register_drawable(self.generate_mesh(self.model.mesh, *args, **kwargs))

    def blocks(self, *args, **kwargs) -> BlockGL:
        return self.register_drawable(self.generate_blocks(self.model.blocks, *args, **kwargs))

    def points(self, *args, **kwargs) -> PointGL:
        return self.register_drawable(self.generate_points(self.model.points, *args, **kwargs))

    def lines(self, *args, **kwargs) -> LineGL:
        return self.register_drawable(self.generate_lines(self.model.lines, *args, **kwargs))

    def tubes(self, *args, **kwargs) -> TubeGL:
        return self.register_drawable(self.generate_tubes(self.model.tubes, *args, **kwargs))

    """
    Load methods by path
    """
    def mesh_by_path(self, path: str, *args, **kwargs) -> MeshGL:
        return self.register_drawable(self.generate_mesh(self.model.mesh_by_path, path, *args, **kwargs))

    def blocks_by_path(self, path: str, *args, **kwargs) -> BlockGL:
        return self.register_drawable(self.generate_blocks(self.model.blocks_by_path, path, *args, **kwargs))

    def points_by_path(self, path: str, *args, **kwargs) -> PointGL:
        return self.register_drawable(self.generate_points(self.model.points_by_path, path, *args, **kwargs))

    def lines_by_path(self, path: str, *args, **kwargs) -> LineGL:
        return self.register_drawable(self.generate_lines(self.model.lines_by_path, path, *args, **kwargs))

    def tubes_by_path(self, path: str, *args, **kwargs) -> TubeGL:
        return self.register_drawable(self.generate_tubes(self.model.tubes_by_path, path, *args, **kwargs))

    def meshes_by_folder_path(self, path: str, *args, **kwargs) -> list:
        return self.register_folder(path, self.mesh_by_path, *args, **kwargs)

    def blocks_by_folder_path(self, path: str, *args, **kwargs) -> list:
        return self.register_folder(path, self.blocks_by_path, *args, **kwargs)

    def points_by_folder_path(self, path: str, *args, **kwargs) -> list:
        return self.register_folder(path, self.points_by_path, *args, **kwargs)

    def lines_by_folder_path(self, path: str, *args, **kwargs) -> list:
        return self.register_folder(path, self.lines_by_path, *args, **kwargs)

    def tubes_by_folder_path(self, path: str, *args, **kwargs) -> list:
        return self.register_folder(path, self.tubes_by_path, *args, **kwargs)

    """
    Export methods
    """
    def _export_element(self, exporter, path: str, _id: int) -> None:
        try:
            exporter(path, _id)
            self.signal_export_success.emit(_id)
        except Exception:
            self.signal_export_failure.emit()

    def export_mesh(self, path, _id) -> None:
        self._export_element(self.model.export_mesh, path, _id)

    def export_blocks(self, path, _id) -> None:
        self._export_element(self.model.export_blocks, path, _id)

    def export_points(self, path, _id) -> None:
        self._export_element(self.model.export_points, path, _id)

    def export_lines(self, path, _id) -> None:
        self._export_element(self.model.export_lines, path, _id)

    def export_tubes(self, path, _id) -> None:
        self._export_element(self.model.export_tubes, path, _id)

    """
    Drawable manipulation
    """
    def get_drawable(self, _id: int):
        return self.drawable_collection.get(_id)

    def show_drawable(self, _id: int) -> None:
        self.get_drawable(_id).show()

    def hide_drawable(self, _id: int) -> None:
        self.get_drawable(_id).hide()

    def update_drawable(self, _id: int) -> None:
        self.makeCurrent()
        self.get_drawable(_id).setup_attributes()
        self.recreate()

    def delete(self, _id: int) -> None:
        if _id < 0:
            return

        self.makeCurrent()
        self.model.delete(_id)
        self.drawable_collection.delete(_id)
        self.signal_file_modified.emit()

    def recreate(self) -> None:
        self.axis_collection.recreate()
        self.background_collection.recreate()
        self.drawable_collection.recreate()
        self.update()

    def get_all_ids(self) -> list:
        return self.drawable_collection.get_all_ids()

    def get_all_drawables(self) -> list:
        return self.drawable_collection.get_all_drawables()

    def update_all(self) -> None:
        for _id in self.get_all_ids():
            self.update_drawable(_id)

    def clear(self) -> None:
        for _id in self.get_all_ids():
            self.delete(_id)

    """
    Camera handling
    """
    def camera_at(self, _id: int) -> None:
        self.fit_to_bounds(*self.get_drawable(_id).element.bounding_box)
        self.update()

    def plan_view(self) -> None:
        self.rotation_angle = [0.0, 0.0, 0.0]
        self.update()

    def north_view(self) -> None:
        self.rotation_angle = [270.0, 0.0, 270.0]
        self.update()

    def east_view(self) -> None:
        self.rotation_angle = [270.0, 0.0, 0.0]
        self.update()

    def fit_to_bounds(self, min_bound: np.ndarray, max_bound: np.ndarray) -> None:
        center = (min_bound + max_bound) / 2
        self.rotation_center = center
        self.camera_position = center

        # Put the camera in a position that allow us to see between the boundaries.
        # In perspective projection, we need a clever trigonometric calculation.
        md = np.max(np.diff([min_bound, max_bound], axis=0))
        aspect = self.width() / self.height()
        fov_rad = self.fov * np.pi / 180.0
        dist = (md / np.tan(fov_rad / 2) + md) / 2.0
        camera_shift = dist * max(1.0, 1.0 / aspect)

        # But in orthographic projection, it's more direct.
        if self.projection_mode == 'orthographic':
            dist = md / 2
            camera_shift = dist * max(1.0, aspect)

        self.zCameraPos += 1.1 * camera_shift
        self.update()

    def fit_to_screen(self) -> None:
        drawables = [d for d in self.get_all_drawables() if d.is_visible]

        if len(drawables) == 0:
            return

        min_all = np.inf * np.ones(3)
        max_all = -np.inf * np.ones(3)

        for drawable in drawables:
            min_bound, max_bound = drawable.element.bounding_box
            min_all = np.min((min_all, min_bound), axis=0)
            max_all = np.max((max_all, max_bound), axis=0)

        self.fit_to_bounds(min_all, max_all)
        self.update()

    """
    Internal methods
    """
    def initializeGL(self) -> None:
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glEnable(GL_VERTEX_PROGRAM_POINT_SIZE)
        glEnable(GL_POINT_SPRITE)
        glDisable(GL_CULL_FACE)
        glEnable(GL_DEPTH_TEST)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def paintGL(self) -> None:
        # Clear screen
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        self.world.setToIdentity()
        self.camera.setToIdentity()
        self.proj.setToIdentity()

        # Translate by rotation center (world position)
        self.world.translate(*self.rotation_center)

        # Allow rotation of the world
        self.world.rotate(self.xCenterRot, 1.0, 0.0, 0.0)
        self.world.rotate(self.yCenterRot, 0.0, 1.0, 0.0)
        self.world.rotate(self.zCenterRot, 0.0, 0.0, 1.0)

        # Restore world
        self.world.translate(*-self.rotation_center)

        # Translate the camera
        self.camera.translate(*-self.camera_position)

        # Project (Perspective/Orthographic)
        self.resizeGL(self.width(), self.height())

        # Draw every GLDrawable (meshes, blocks, points, etc)
        glEnable(GL_BLEND)
        self.background_collection.draw(self.proj, self.camera, self.world)
        self.drawable_collection.draw(self.proj, self.camera, self.world)
        self.axis_collection.draw(self.proj, self.camera, self.world)

        # Tick FPS counter
        self.fps_counter.tick(callback=self.signal_fps_updated.emit)

    def resizeGL(self, w: float, h: float) -> None:
        if self.projection_mode == 'perspective':
            self.proj.perspective(self.fov, (w / h), 1.0, 100000.0)
        elif self.projection_mode == 'orthographic':
            z = self.off_center[2]
            aspect = w / h
            self.proj.ortho(-z, z, -z / aspect, z / aspect, 0.0, 100000.0)

    """
    Utilities
    """
    def get_pixmap(self) -> QPixmap:
        pixmap = QPixmap(self.size())
        self.render(pixmap, QPoint(), QRegion(self.rect()))
        return pixmap

    def take_screenshot(self, save_path=None) -> None:
        if save_path:
            self.get_pixmap().save(save_path)

    def screen_to_ndc(self, _x, _y, _z) -> np.ndarray:
        # Click at bottom-left of screen => (-1.0, -1.0, z)
        # Click at top-right of screen => (1.0, 1.0, z)
        # But we can't really know where's z, so we just return 1.0
        x = (2.0 * _x / self.width()) - 1.0
        y = 1.0 - (2.0 * _y / self.height())
        z = 1.0
        return np.array([x, y, z])

    def unproject(self, _x, _y, _z, model, view, proj) -> np.ndarray:
        # Adapted from http://antongerdelan.net/opengl/raycasting.html
        x, y, z = self.screen_to_ndc(_x, _y, _z)

        # We'd use `QVector4D(x, y, -1.0, 1.0)`, but PySide2
        # hasn't implemented QMatrix4x4 * QVector4D yet.
        vector = [x, y, -1.0, 1.0]
        temp_matrix = QMatrix4x4(*[e for e in vector for _ in range(4)])

        ray_eye = (proj.inverted()[0] * temp_matrix).column(0)
        ray_eye = QVector4D(ray_eye.x(), ray_eye.y(), -1.0, 0.0)

        # We'd use `ray_eye`, but PySide2
        # hasn't implemented QMatrix4x4 * QVector4D yet.
        vector = [ray_eye.x(), ray_eye.y(), ray_eye.z(), ray_eye.w()]
        temp_matrix = QMatrix4x4(*[e for e in vector for _ in range(4)])

        ray_world = ((view * model).inverted()[0] * temp_matrix).column(0).toVector3D()
        ray = ray_world.normalized()
        return np.array([ray.x(), ray.y(), ray.z()])

    @staticmethod
    def get_origin(model: QMatrix4x4, view: QMatrix4x4) -> np.ndarray:
        origin = (view * model).inverted()[0].column(3).toVector3D()
        return np.array([origin.x(), origin.y(), origin.z()])

    def ray_from_click(self, x: float, y: float, z: float) -> np.ndarray:
        # Perspective projection is straightforward
        if self.projection_mode == 'perspective':
            return self.unproject(x, y, z, self.world, self.camera, self.proj)

        # Orthographic projection "forces" a click in the center
        return self.unproject(self.width() / 2, self.height() / 2, z,
                              self.world, self.camera, self.proj)

    def origin_from_click(self, x: float, y: float, z: float) -> np.ndarray:
        # Perspective projection is straightforward
        if self.projection_mode == 'perspective':
            return self.get_origin(self.world, self.camera)

        # Orthographic projection needs a bit more of vector arithmetic.
        # A click in the center of the screen gives us the perfect ray,
        # by recycling the perspective's un-project method.

        # But if we don't click in the exact center of screen,
        # we need to trick the origin calculation.
        ndc = self.screen_to_ndc(x, y, z)
        aspect = self.width() / self.height()
        aspect_bias = np.array([1.0, 1.0 / aspect, 0.0])

        # The idea is to strategically move the camera by an offset, so that
        # clicking anywhere in the screen will give us the same result as
        # clicking at the exact center of the screen from a shifted camera.
        offset = self.off_center[2] * ndc * aspect_bias

        # Hack to get the origin when the click is not exactly at center of screen.
        self.camera.translate(*-offset)
        origin = self.get_origin(self.world, self.camera)
        self.camera.translate(*offset)

        return origin

    def slice_meshes(self, origin: np.ndarray, plane_normal: np.ndarray) -> None:
        # Slicing all *visible* meshes
        meshes = [m.element for m in self.drawable_collection.filter(MeshGL) if m.is_visible]

        results = self.model.slice_meshes(origin, plane_normal, meshes)
        self.signal_mesh_sliced.emit(results)

    def slice_blocks(self, origin: np.ndarray, plane_normal: np.ndarray) -> None:
        # Slicing all blocks
        blocks = [m.element for m in self.drawable_collection.filter(BlockGL)]

        results = self.model.slice_blocks(origin, plane_normal, blocks)
        self.signal_blocks_sliced.emit(results)

    def slice_drawables(self, origin: np.ndarray, plane_normal: np.ndarray) -> None:
        self.slice_meshes(origin, plane_normal)
        self.slice_blocks(origin, plane_normal)

    def slice_from_rays(self, origin_list: list, ray_list: list) -> None:
        # A plane is created from `origin` and `ray_list`.
        # In perspective projection, the origin is the same.
        origin = origin_list[0]

        if self.projection_mode == 'perspective':
            # Perspective: Same origins, different rays
            plane_normal = np.cross(*ray_list)
        else:
            # Orthographic: Same rays, different origins
            origin_diff = np.diff(origin_list, axis=0)[0]
            plane_normal = np.cross(ray_list[0], origin_diff)

        plane_normal /= np.linalg.norm(plane_normal)

        # Slice drawables
        self.slice_drawables(origin, plane_normal)

        # Auto-exit the slice mode for now
        self.set_normal_mode()

    def measure_from_rays(self, origin_list: list, ray_list: list) -> None:
        meshes = [m.element for m in self.drawable_collection.filter(MeshGL) if m.is_visible]
        results = self.model.measure_from_rays(origin_list, ray_list, meshes)

        self.signal_mesh_distances.emit(results)

    def detect_mesh_intersection(self, x: float, y: float, z: float) -> None:
        ray = self.ray_from_click(x, y, z)
        origin = self.origin_from_click(x, y, z)
        elements = [m.element for m in self.drawable_collection.filter(MeshGL) if m.is_visible]

        attributes_list = []

        for mesh in elements:
            intersections = mesh.intersect_with_ray(origin, ray)
            closest_point = utils.closest_point_to(origin, intersections)
            if closest_point is not None:
                attributes = {**mesh.attributes,
                              'intersections': intersections,
                              'closest_point': closest_point,
                              }
                attributes_list.append(attributes)

        # Emit signal with clicked mesh
        self.signal_mesh_clicked.emit(attributes_list)

    """
    Controller
    """
    def add_controller(self, mode, mode_name: str) -> None:
        self.controllers[mode_name] = mode

    def set_controller(self, mode_name: str) -> None:
        self.current_mode = self.controllers.get(mode_name)()
        self.signal_mode_updated.emit(self.current_mode.name)
        self.update()

    def set_normal_mode(self) -> None:
        self.set_controller('normal')

    def set_detection_mode(self) -> None:
        self.set_controller('detection')

    def set_slice_mode(self) -> None:
        self.set_controller('slice')

    def set_measurement_mode(self) -> None:
        self.set_controller('measurement')

    """
    Events (dependent on current controller)
    """
    def mouseMoveEvent(self, event, *args, **kwargs) -> None:
        self.current_mode.mouseMoveEvent(event, self)
        self.update()

    def mousePressEvent(self, event, *args, **kwargs) -> None:
        self.current_mode.mousePressEvent(event, self)
        self.update()

    def mouseDoubleClickEvent(self, event, *args, **kwargs) -> None:
        self.current_mode.mouseDoubleClickEvent(event, self)
        self.update()

    def mouseReleaseEvent(self, event, *args, **kwargs) -> None:
        self.current_mode.mouseReleaseEvent(event, self)
        self.update()

    def wheelEvent(self, event, *args, **kwargs) -> None:
        self.current_mode.wheelEvent(event, self)
        self.update()

    def dragEnterEvent(self, event, *args, **kwargs) -> None:
        if event.mimeData().hasText():
            event.acceptProposedAction()

    def dropEvent(self, event, *args, **kwargs) -> None:
        for url in event.mimeData().urls():
            path = url.toLocalFile()

            if QFileInfo(path).isDir():
                self.meshes_by_folder_path(path)
            else:
                self.mesh_by_path(path)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtGui import QColor
from qtpy.QtWidgets import QColorDialog


class ColorDialog(QColorDialog):
    def __init__(self, parent=None, element=None):
        QColorDialog.__init__(self, parent)
        self.setOption(self.ShowAlphaChannel)
        self.setOption(self.DontUseNativeDialog)

        self.setWindowTitle(f'{self.windowTitle()} ({element.name}.{element.extension})')
        self.setCurrentColor(QColor.fromRgb(*[int(255 * x) for x in element.rgba.tolist()]))



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtGui import QMouseEvent
from .mode import Mode


class DetectionMode(Mode):
    def __init__(self):
        super().__init__()
        self.name = 'Detection Mode'

    def mousePressEvent(self, event: QMouseEvent, viewer) -> None:
        viewer.detect_mesh_intersection(event.pos().x(), event.pos().y(), 1.0)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import time


class FPSCounter:
    def __init__(self, resolution: float = 1.0):
        self.start_time = time.time()
        self.resolution = resolution
        self.counter = 0

    def tick(self, callback=lambda *args: None) -> None:
        self.counter += 1
        if time.time() - self.start_time > self.resolution:
            diff = (time.time() - self.start_time)
            callback(self.counter / diff if diff else 0.0)
            self.counter = 0
            self.start_time = time.time()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
import pathlib

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog
from qtpy.QtWidgets import QTableWidgetItem
from .tools import uic


class PropertiesDialog(QDialog):
    def __init__(self, parent=None, element=None):
        QDialog.__init__(self, parent)
        uic.loadUi(f'{pathlib.Path(__file__).parent}/UI/propertiesdialog.ui', self)

        # Avoids the QObject::startTimer warning (maybe)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowTitle(f'Set properties ({element.name}.{element.extension})')

        # Fill headers and set current ones
        self.fill_headers(element.all_headers)
        self.current_headers = element.headers

        # Fill properties in QTableWidget
        self.tableWidget_properties.setRowCount(len(element.customizable_properties))
        self.tableWidget_properties.setVerticalHeaderLabels(element.customizable_properties)

        for i, k in enumerate(element.customizable_properties):
            v = getattr(element, k)
            text = str(v.tolist()) if type(v) is np.ndarray else str(v)
            self.tableWidget_properties.setItem(i, 0, QTableWidgetItem(text))

    def fill_headers(self, headers: list) -> None:
        for i in headers:
            self.comboBox_x.addItem(i)
            self.comboBox_y.addItem(i)
            self.comboBox_z.addItem(i)
            self.comboBox_values.addItem(i)

    def has_altered_coordinates(self, element) -> bool:
        altered = False
        for e, c in list(zip(element.headers, self.current_headers))[:3]:
            altered |= not (e == c)

        return altered

    @property
    def current_headers(self) -> list:
        return [self.comboBox_x.currentText(),
                self.comboBox_y.currentText(),
                self.comboBox_z.currentText(),
                self.comboBox_values.currentText()]

    @current_headers.setter
    def current_headers(self, headers: list) -> None:
        self.comboBox_x.setCurrentIndex(self.comboBox_x.findText(headers[0]))
        self.comboBox_y.setCurrentIndex(self.comboBox_y.findText(headers[1]))
        self.comboBox_z.setCurrentIndex(self.comboBox_z.findText(headers[2]))
        self.comboBox_values.setCurrentIndex(self.comboBox_values.findText(headers[3]))



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np


class Element:
    __slots__ = ['_data', '_properties', '_metadata']

    def __init__(self, *args, **kwargs):
        """
        Element is a class designed with the following idea:
        (replace list for numpy arrays in the implementation)

        {
            'data': {
                'x': list[float],
                'y': list[float],
                'z': list[float]
            }
            'properties': {
                'color': list[float],
                'alpha': float
            },
            'metadata': {
                'id': int,
                'name': str,
                'extension': str
            }
        }
        """
        # Base data
        self._data: dict = {}
        self._properties: dict = {}
        self._metadata: dict = {}

        self._initialize(*args, **kwargs)

    def _initialize(self, *args, **kwargs) -> None:
        self._fill_element(*args, **kwargs)
        self._fill_metadata(*args, **kwargs)
        self._fill_properties(*args, **kwargs)
        self._check_integrity()

    """
    Element filling
    """
    def _fill_element(self, *args, **kwargs) -> None:
        msg = f'Data must contain ["x", "y", "z"], "vertices" or "data", got {list(kwargs.keys())}.'
        if 'data' in kwargs.keys():
            self._fill_as_data(*args, **kwargs)
        elif 'vertices' in kwargs.keys():
            self._fill_as_vertices(*args, **kwargs)
        elif 'x' in kwargs.keys() and 'y' in kwargs.keys() and 'z' in kwargs.keys():
            self._fill_as_xyz(*args, **kwargs)
        else:
            raise KeyError(msg)

    def _fill_as_data(self, *args, **kwargs) -> None:
        data = kwargs.get('data', {})
        if 'vertices' in data.keys():
            self.vertices = data.get('vertices', [])
        else:
            self.x = data.get('x', [])
            self.y = data.get('y', [])
            self.z = data.get('z', [])

    def _fill_as_vertices(self, *args, **kwargs) -> None:
        self.vertices = kwargs.get('vertices', [])

    def _fill_as_xyz(self, *args, **kwargs) -> None:
        self.x = kwargs.get('x', [])
        self.y = kwargs.get('y', [])
        self.z = kwargs.get('z', [])

    def _fill_properties(self, *args, **kwargs) -> None:
        self.color = kwargs.get('color', np.random.rand(3))
        self.alpha = kwargs.get('alpha', 1.0)

    def _fill_metadata(self, *args, **kwargs) -> None:
        self.name = kwargs.get('name')
        self.extension = kwargs.get('extension')
        self.id = kwargs.get('id', -1)

    def _check_integrity(self) -> None:
        msg = f'Coordinates have different lengths: ({self.x.size}, {self.y.size}, {self.z.size})'
        if not (self.x.size == self.y.size == self.z.size):
            raise ValueError(msg)

    """
    Main accessors
    """
    @property
    def data(self) -> dict:
        return self._data

    @property
    def properties(self) -> dict:
        return self._properties

    @property
    def metadata(self) -> dict:
        return self._metadata

    @property
    def attributes(self) -> dict:
        return {**self.properties, **self.metadata}

    @data.setter
    def data(self, _data: dict) -> None:
        self._data = _data

    @properties.setter
    def properties(self, _properties: dict) -> None:
        for k, v in _properties.items():
            self._properties[k] = v

    """
    Data
    """
    @property
    def x(self) -> np.ndarray:
        return self.data.get('x')

    @property
    def y(self) -> np.ndarray:
        return self.data.get('y')

    @property
    def z(self) -> np.ndarray:
        return self.data.get('z')

    @property
    def vertices(self) -> np.ndarray:
        return np.column_stack((self.x, self.y, self.z))

    @x.setter
    def x(self, val: list) -> None:
        self.data['x'] = np.array(val, np.float32)

    @y.setter
    def y(self, val: list) -> None:
        self.data['y'] = np.array(val, np.float32)

    @z.setter
    def z(self, val: list) -> None:
        self.data['z'] = np.array(val, np.float32)

    @vertices.setter
    def vertices(self, vertex_list: list) -> None:
        self.x, self.y, self.z = np.array(vertex_list).T

    """
    Properties
    """
    @property
    def customizable_properties(self) -> list:
        return ['color', 'alpha']

    @property
    def exportable_properties(self) -> list:
        return ['color', 'alpha']

    @property
    def color(self) -> np.array:
        return self.properties.get('color')

    @property
    def alpha(self) -> float:
        return self.properties.get('alpha')

    @property
    def rgba(self) -> np.ndarray:
        return np.append(self.color, self.alpha)

    @color.setter
    def color(self, val: list) -> None:
        self.properties['color'] = np.array(val)

    @alpha.setter
    def alpha(self, val: float) -> None:
        self.properties['alpha'] = val

    @rgba.setter
    def rgba(self, val: list) -> None:
        self.color = np.array(val)[:3]
        self.alpha = val[-1]

    """
    Utilities
    """
    @property
    def centroid(self) -> np.ndarray:
        return self.vertices.mean(axis=0)

    @property
    def center(self) -> np.ndarray:
        return np.sum(self.bounding_box, axis=0) / 2

    @property
    def bounding_box(self) -> tuple:
        # We could return self.vertices.min(axis=0), self.vertices.max(axis=0),
        # but I found this to be noticeable faster (speedup of 5.2 approx.)
        return (np.array([self.x.min(), self.y.min(), self.z.min()]),
                np.array([self.x.max(), self.y.max(), self.z.max()]))

    """
    Metadata
    """
    @property
    def id(self) -> int:
        return self.metadata.get('id')

    @property
    def name(self) -> str:
        return self.metadata.get('name')

    @property
    def extension(self) -> str:
        return self.metadata.get('extension')

    @id.setter
    def id(self, _id: int) -> None:
        self.metadata['id'] = _id

    @name.setter
    def name(self, _name: str) -> None:
        self.metadata['name'] = _name

    @extension.setter
    def extension(self, _extension: str) -> None:
        self.metadata['extension'] = _extension



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtWidgets import QToolBar
from .actioncollection import ActionCollection


class ToolBar(QToolBar):
    def __init__(self, parent=None):
        QToolBar.__init__(self, parent)
        self.setAcceptDrops(True)

        self.action_collection = ActionCollection(self)
        self.add_actions()

    def add_actions(self) -> None:
        self.addAction(self.action_collection.action_show_tree)
        self.addSeparator()
        self.addAction(self.action_collection.action_plan_view)
        self.addAction(self.action_collection.action_north_view)
        self.addAction(self.action_collection.action_east_view)
        self.addAction(self.action_collection.action_fit_to_screen)
        self.addSeparator()
        self.addAction(self.action_collection.action_autofit_to_screen)
        self.addAction(self.action_collection.action_turbo_rendering)
        self.addSeparator()
        self.addAction(self.action_collection.action_take_screenshot)

    def connect_tree(self, tree) -> None:
        self.action_collection.connect_tree(tree)

    def connect_main_widget(self, widget) -> None:
        self.action_collection.connect_main_widget(widget)

    def connect_viewer(self, viewer) -> None:
        self.action_collection.connect_viewer(viewer)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from collections import OrderedDict
from .glprogramcollection import GLProgramCollection
from ..drawables.gldrawable import GLDrawable


class GLCollection:
    def __init__(self, viewer=None):
        self._viewer = viewer
        self._programs = GLProgramCollection()
        self._collection = OrderedDict()
        self.needs_update = True

    """
    Drawable collection handlers
    """
    def add(self, drawable: GLDrawable) -> None:
        self._collection[drawable.id] = drawable
        self.needs_update = True

    def get(self, _id: int) -> GLDrawable:
        return self._collection.get(_id)

    def get_all_ids(self) -> list:
        return list(self._collection.keys())

    def get_all_drawables(self) -> list:
        return list(self._collection.values())

    def delete(self, _id: int) -> None:
        drawable = self._collection.pop(_id)
        drawable.cleanup()

    def filter(self, drawable_type: type) -> list:
        # The copy avoids RuntimeError: OrderedDict mutated during iteration
        return [x for x in self._collection.copy().values() if type(x) is drawable_type]

    def clear(self) -> None:
        self._collection.clear()

    def size(self) -> int:
        return len(self._collection)

    @property
    def last_id(self) -> int:
        # bool(dict) evaluates to False if the dictionary is empty
        return list(self._collection.keys())[-1] if bool(self._collection) else -1

    """
    ShaderProgram collection handlers
    """
    def associate(self, program, association) -> None:
        """
        :param program: AxisProgram, for example
        :param association: lambda
        :return: None
        """
        self._programs.associate(program, association)

    def recreate(self) -> None:
        self.needs_update = True
        for program in self._programs.get_programs():
            program.recreate()

    def update_drawables(self) -> None:
        # Update shader program so that it knows what to draw
        if self.needs_update:
            for program, drawable_retriever in self._programs.get_pairs():
                program.set_drawables([d for d in drawable_retriever() if d.is_visible])
            self.needs_update = False

    def update_matrices(self, proj, view, model) -> None:
        # Update matrices so that it knows where it's looking
        for program in self._programs.get_programs():
            if len(program.drawables) > 0:
                program.setup()
                program.bind()
                program.update_uniform('proj_matrix', proj)
                program.update_uniform('model_view_matrix', view * model)

    """
    Drawing methods
    """
    def draw_opaques(self) -> None:
        for program in self._programs.get_programs():
            if len(program.opaques) > 0:
                program.bind()
                program.draw()

    def draw_transparents(self) -> None:
        for program in self._programs.get_programs():
            if len(program.transparents) > 0:
                program.bind()
                program.redraw()

    def draw(self, proj_matrix, view_matrix, model_matrix) -> None:
        self.update_drawables()
        self.update_matrices(proj_matrix, view_matrix, model_matrix)
        self.draw_opaques()
        self.draw_transparents()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import pathlib

from qtpy.QtCore import Qt
from qtpy.QtCore import QFileInfo
from qtpy.QtCore import QSettings
from qtpy.QtCore import QThreadPool
from qtpy.QtWidgets import QFileDialog
from qtpy.QtWidgets import QMainWindow

from .helpdialog import HelpDialog
from .aboutdialog import AboutDialog
from .threadworker import ThreadWorker
from .iconcollection import IconCollection

from .tools import uic


class MainWindow(QMainWindow):
    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)
        uic.loadUi(f'{pathlib.Path(__file__).parent}/UI/mainwindow.ui', self)
        self.resize(1000, 500)

        self.setWindowTitle('BlastSight')
        self.setWindowIcon(IconCollection.get('blastsight.png'))
        self.toolbar.setWindowTitle('Toolbar')
        self.treeWidget.enable_exportability(True)

        self.setFocusPolicy(Qt.StrongFocus)
        self.setAcceptDrops(True)
        self.statusBar.showMessage('Ready')

        # Attributes
        self.settings = QSettings('BlastSight', application='blastsight', parent=self)
        self.filters_dict = {
            'mesh': 'Mesh Files (*.dxf *.off *.h5m);;'
                    'DXF Files (*.dxf);;'
                    'OFF Files (*.off);;'
                    'H5M Files (*.h5m);;'
                    'All Files (*.*)',
            'block': 'Block Files (*.csv *.h5p *.out);;'
                     'CSV Files (*.csv);;'
                     'H5P Files (*.h5p);;'
                     'GSLib Files (*.out);;'
                     'All Files (*.*)',
            'point': 'Point Files (*.csv *.h5p *.out);;'
                     'CSV Files (*.csv);;'
                     'H5P Files (*.h5p);;'
                     'GSLib Files (*.out);;'
                     'All Files (*.*)',
            'line': 'Line Files (*.csv *.dxf);;'
                    'CSV Files (*.csv);;'
                    'DXF Files (*.csv);;'
                    'All Files (*.*)',
            'tube': 'Tube Files (*.csv *.dxf);;'
                    'CSV Files (*.csv);;'
                    'DXF Files (*.csv);;'
                    'All Files (*.*)',
        }

        # Extra actions
        self.toolbar.insertAction(self.toolbar.action_collection.action_plan_view,
                                  self.toolbar.action_collection.action_camera_properties)
        self.toolbar.addAction(self.toolbar.action_collection.action_quit)
        self.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)

        self.generate_menubar()
        self.connect_actions()

        # Set auto-fit as True when starting the app
        if not self.toolbar.action_collection.action_autofit_to_screen.isChecked():
            self.toolbar.action_collection.action_autofit_to_screen.trigger()

        # self.title = self.windowTitle()
        # self.viewer.signal_fps_updated.connect(lambda x: self.setWindowTitle(f'{self.title} (FPS: {x:.1f})'))

    def generate_menubar(self) -> None:
        self.menu_File.addAction(self.toolbar.action_collection.action_load_mesh)
        self.menu_File.addAction(self.toolbar.action_collection.action_load_blocks)
        self.menu_File.addAction(self.toolbar.action_collection.action_load_points)
        self.menu_File.addAction(self.toolbar.action_collection.action_load_lines)
        self.menu_File.addAction(self.toolbar.action_collection.action_load_tubes)
        self.menu_File.addSeparator()
        self.menu_File.addAction(self.toolbar.action_collection.action_load_mesh_folder)
        self.menu_File.addAction(self.toolbar.action_collection.action_load_blocks_folder)
        self.menu_File.addAction(self.toolbar.action_collection.action_load_points_folder)
        self.menu_File.addAction(self.toolbar.action_collection.action_load_lines_folder)
        self.menu_File.addAction(self.toolbar.action_collection.action_load_tubes_folder)
        self.menu_File.addSeparator()
        self.menu_File.addAction(self.toolbar.action_collection.action_quit)

        self.menu_View.addAction(self.toolbar.action_collection.action_camera_properties)
        self.menu_View.addAction(self.toolbar.action_collection.action_plan_view)
        self.menu_View.addAction(self.toolbar.action_collection.action_north_view)
        self.menu_View.addAction(self.toolbar.action_collection.action_east_view)
        self.menu_View.addAction(self.toolbar.action_collection.action_fit_to_screen)
        self.menu_View.addSeparator()
        self.menu_View.addAction(self.toolbar.action_collection.action_autofit_to_screen)
        self.menu_View.addAction(self.toolbar.action_collection.action_turbo_rendering)
        self.menu_View.addSeparator()
        self.menu_View.addAction(self.toolbar.action_collection.action_perspective_projection)
        self.menu_View.addAction(self.toolbar.action_collection.action_orthographic_projection)
        self.menu_View.addSeparator()
        self.menu_View.addAction(self.toolbar.action_collection.action_take_screenshot)

        self.menu_Tools.addAction(self.toolbar.action_collection.action_slice_mode)
        self.menu_Tools.addAction(self.toolbar.action_collection.action_detection_mode)
        self.menu_Tools.addAction(self.toolbar.action_collection.action_measurement_mode)
        self.menu_Tools.addSeparator()
        self.menu_Tools.addAction(self.toolbar.action_collection.action_normal_mode)

        self.menu_Help.addAction(self.toolbar.action_collection.action_help)
        self.menu_Help.addAction(self.toolbar.action_collection.action_about)

    def connect_actions(self) -> None:
        # Auto-connections
        self.toolbar.connect_tree(self.treeWidget)  # Handles tree.show()
        self.toolbar.connect_viewer(self.viewer)  # Handles Plan/North/East/Fit/Camera/Turbo/Screenshot
        self.toolbar.connect_main_widget(self)  # Handles Exit
        self.treeWidget.connect_viewer(self.viewer)  # Handles signals from viewer

        # File
        self.toolbar.action_collection.action_load_mesh.triggered.connect(self.dialog_load_mesh)
        self.toolbar.action_collection.action_load_blocks.triggered.connect(self.dialog_load_blocks)
        self.toolbar.action_collection.action_load_points.triggered.connect(self.dialog_load_points)
        self.toolbar.action_collection.action_load_lines.triggered.connect(self.dialog_load_lines)
        self.toolbar.action_collection.action_load_tubes.triggered.connect(self.dialog_load_tubes)

        self.toolbar.action_collection.action_load_mesh_folder.triggered.connect(self.dialog_load_mesh_folder)
        self.toolbar.action_collection.action_load_blocks_folder.triggered.connect(self.dialog_load_blocks_folder)
        self.toolbar.action_collection.action_load_points_folder.triggered.connect(self.dialog_load_points_folder)
        self.toolbar.action_collection.action_load_lines_folder.triggered.connect(self.dialog_load_lines_folder)
        self.toolbar.action_collection.action_load_tubes_folder.triggered.connect(self.dialog_load_tubes_folder)

        # View
        self.toolbar.action_collection.action_show_tree.triggered.connect(self.dockWidget.show)
        self.toolbar.action_collection.action_perspective_projection.triggered.connect(self.viewer.perspective_projection)
        self.toolbar.action_collection.action_orthographic_projection.triggered.connect(self.viewer.orthographic_projection)

        # Tools
        self.toolbar.action_collection.action_detection_mode.triggered.connect(self.slot_detection_mode)
        self.toolbar.action_collection.action_slice_mode.triggered.connect(self.slot_slice_mode)
        self.toolbar.action_collection.action_measurement_mode.triggered.connect(self.slot_measurement_mode)
        self.toolbar.action_collection.action_normal_mode.triggered.connect(self.slot_normal_mode)

        # Help
        self.toolbar.action_collection.action_help.triggered.connect(self.slot_help)
        self.toolbar.action_collection.action_about.triggered.connect(self.slot_about)

        # Viewer signals
        self.viewer.signal_fps_updated.connect(self.print_fps)
        self.viewer.signal_mode_updated.connect(self.slot_mode_updated)
        self.viewer.signal_mesh_clicked.connect(self.slot_mesh_clicked)
        self.viewer.signal_mesh_distances.connect(self.slot_mesh_distances)
        self.viewer.signal_mesh_sliced.connect(self.slot_mesh_sliced)
        self.viewer.signal_blocks_sliced.connect(self.slot_blocks_sliced)

        self.viewer.signal_load_success.connect(self.slot_element_load_success)
        self.viewer.signal_load_failure.connect(self.slot_element_load_failure)
        self.viewer.signal_export_success.connect(self.slot_element_export_success)
        self.viewer.signal_export_failure.connect(self.slot_element_export_failure)

        # TreeWidget actions
        self.treeWidget.signal_export_mesh.connect(self.dialog_export_mesh)
        self.treeWidget.signal_export_blocks.connect(self.dialog_export_blocks)
        self.treeWidget.signal_export_points.connect(self.dialog_export_points)
        self.treeWidget.signal_export_lines.connect(self.dialog_export_lines)
        self.treeWidget.signal_export_tubes.connect(self.dialog_export_tubes)

    @property
    def last_dir(self) -> str:
        return self.settings.value('last_directory', '.')

    @last_dir.setter
    def last_dir(self, _last_dir: str) -> None:
        self.settings.setValue('last_directory', _last_dir)

    @staticmethod
    def print_fps(fps) -> None:
        print(f'               \r', end='')
        print(f'FPS: {fps:.1f} \r', end='')

    """
    Status bar updates
    """
    def slot_element_load_success(self, _id: int) -> None:
        self.statusBar.showMessage(f'Load successful (id: {_id}).')

    def slot_element_load_failure(self) -> None:
        self.statusBar.showMessage(f'Failed to load.')

    def slot_element_export_success(self, _id: int) -> None:
        self.statusBar.showMessage(f'Export successful (id: {_id}).')

    def slot_element_export_failure(self) -> None:
        self.statusBar.showMessage(f'Failed to export.')

    def slot_mode_updated(self, mode: str) -> None:
        self.statusBar.showMessage(f'Mode: {mode}')

    def slot_mesh_distances(self, distance_dict: dict) -> None:
        self.statusBar.showMessage(f'Distance: {distance_dict.get("distance")}')

    def slot_mesh_clicked(self, mesh_attributes: list) -> None:
        id_list = [attr.get('id', -1) for attr in mesh_attributes]
        self.statusBar.showMessage(f'Detected meshes: {id_list}')

    def slot_mesh_sliced(self, slice_dict: dict) -> None:
        slice_list = slice_dict.get('slices', [])

        for sliced_meshes in slice_list:
            slices = sliced_meshes.get('vertices')
            origin_id = sliced_meshes.get('origin_id')
            mesh = self.viewer.get_drawable(origin_id)

            if 'SLICE' in str(mesh.name):
                continue

            for i, vert_slice in enumerate(slices):
                self.viewer.lines(vertices=vert_slice,
                                  color=mesh.color,
                                  name=f'MESHSLICE_{i}_{mesh.name}',
                                  extension=mesh.extension,
                                  loop=True)

    def slot_blocks_sliced(self, slice_dict: dict) -> None:
        slice_list = slice_dict.get('slices', [])

        for sliced_blocks in slice_list:
            indices = sliced_blocks.get('indices')
            origin_id = sliced_blocks.get('origin_id')
            block = self.viewer.get_drawable(origin_id)

            if 'SLICE' in str(block.name):
                continue

            self.viewer.blocks(vertices=block.vertices[indices],
                               values=block.values[indices],
                               color=block.color[indices],
                               vmin=block.vmin,
                               vmax=block.vmax,
                               colormap=block.colormap,
                               name=f'BLOCKSLICE_{block.name}',
                               extension=block.extension,
                               block_size=block.block_size,
                               alpha=1.0,
                               )

    """
    Common functionality for loading/exporting
    """
    def _threaded_load(self, method: classmethod, path: str, *args, **kwargs) -> None:
        self.statusBar.showMessage('Loading...')

        worker = ThreadWorker(method, path, *args, **kwargs)
        QThreadPool.globalInstance().start(worker)

    def _threaded_export(self, method: classmethod, path: str, _id: int) -> None:
        self.statusBar.showMessage('Exporting...')

        worker = ThreadWorker(method, path, _id)
        QThreadPool.globalInstance().start(worker)

    def _dialog_load_element(self, method: classmethod, hint: str, *args, **kwargs) -> None:
        (paths, selected_filter) = QFileDialog.getOpenFileNames(
            parent=self,
            directory=self.last_dir,
            filter=self.filters_dict.get(hint))

        path_list = [p for p in paths if p != '']
        for path in sorted(path_list):
            self._threaded_load(method, path, *args, **kwargs)
            self.last_dir = QFileInfo(path).absoluteDir().absolutePath()

    def _dialog_load_folder(self, method: classmethod, hint: str, *args, **kwargs) -> None:
        path = QFileDialog.getExistingDirectory(
            parent=self,
            directory=self.last_dir,
            options=QFileDialog.ShowDirsOnly)

        # Execute method
        self._threaded_load(method, path, *args, **kwargs)
        self.last_dir = path

    def _dialog_export_element(self, _id: int, filters: str, method: classmethod) -> None:
        (path, selected_filter) = QFileDialog.getSaveFileName(
            parent=self,
            directory=self.viewer.get_drawable(_id).element.name,
            filter=filters)

        if path != '':
            # Execute method
            self._threaded_export(method, path, _id)

    """
    Slots for loading files
    """
    def dialog_load_mesh(self) -> None:
        self._dialog_load_element(method=self.viewer.mesh_by_path, hint='mesh')

    def dialog_load_blocks(self) -> None:
        self._dialog_load_element(method=self.viewer.blocks_by_path, hint='block')

    def dialog_load_points(self) -> None:
        self._dialog_load_element(method=self.viewer.points_by_path, hint='point')

    def dialog_load_lines(self) -> None:
        self._dialog_load_element(method=self.viewer.lines_by_path, hint='line')

    def dialog_load_tubes(self) -> None:
        self._dialog_load_element(method=self.viewer.tubes_by_path, hint='tube')

    def dialog_load_mesh_folder(self) -> None:
        self._dialog_load_folder(method=self.viewer.meshes_by_folder_path, hint='mesh')

    def dialog_load_blocks_folder(self) -> None:
        self._dialog_load_folder(method=self.viewer.blocks_by_folder_path, hint='block')

    def dialog_load_points_folder(self) -> None:
        self._dialog_load_folder(method=self.viewer.points_by_folder_path, hint='point')

    def dialog_load_lines_folder(self) -> None:
        self._dialog_load_folder(method=self.viewer.lines_by_folder_path, hint='line')

    def dialog_load_tubes_folder(self) -> None:
        self._dialog_load_folder(method=self.viewer.tubes_by_folder_path, hint='tube')

    """
    Slots for exporting files
    """
    def dialog_export_mesh(self, _id: int) -> None:
        self._dialog_export_element(_id=_id,
                                    filters='BlastSight mesh (*.h5m);;',
                                    method=self.viewer.export_mesh)

    def dialog_export_blocks(self, _id: int) -> None:
        self._dialog_export_element(_id=_id,
                                    filters='BlastSight blocks (*.h5p);;',
                                    method=self.viewer.export_blocks)

    def dialog_export_points(self, _id: int) -> None:
        self._dialog_export_element(_id=_id,
                                    filters='BlastSight points (*.h5p);;',
                                    method=self.viewer.export_points)

    def dialog_export_lines(self, _id: int) -> None:
        self._dialog_export_element(_id=_id,
                                    filters='BlastSight lines (*.csv);;',
                                    method=self.viewer.export_lines)

    def dialog_export_tubes(self, _id: int) -> None:
        self._dialog_export_element(_id=_id,
                                    filters='BlastSight tubes (*.csv);;',
                                    method=self.viewer.export_tubes)

    """
    Slots for modifying controller modes
    """
    def slot_normal_mode(self) -> None:
        self.viewer.set_normal_mode()

    def slot_detection_mode(self) -> None:
        self.viewer.set_detection_mode()

    def slot_slice_mode(self) -> None:
        self.viewer.set_slice_mode()

    def slot_measurement_mode(self) -> None:
        self.viewer.set_measurement_mode()

    """
    Slots for showing help/about dialogs
    """
    def slot_help(self) -> None:
        HelpDialog(self).exec_()

    def slot_about(self) -> None:
        AboutDialog(self).exec_()

    """
    Events pass-through
    """
    def dragEnterEvent(self, event, *args, **kwargs) -> None:
        self.viewer.dragEnterEvent(event, *args, **kwargs)

    def dropEvent(self, event, *args, **kwargs) -> None:
        self.viewer.dropEvent(event, *args, **kwargs)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
from .dfelement import DFElement


class PointElement(DFElement):
    def __init__(self, *args, **kwargs):
        """
        PointElement is a class inheriting from DFElement.

        {
            'data': {
                'x': list[float],
                'y': list[float],
                'z': list[float],
                'values: list[float],
            },
            'datasets': {
                'size': list[float], (Each point has its own size in 1D)
                'color': list[list[float]] (Optional, auto-generated from 'values' if None)
            },
            'properties': {
                'headers': list[str],
                'vmin': float,
                'vmax': float,
                'alpha': float,
                'colormap': str (Optional, used from 'values')
            }
            'metadata': {
                'id': int,
                'name': str or None,
                'extension': str or None
            }
        }

        In a PointElement, each point has its own 1D size, and each block has its own color.
        If the user didn't specify colors, they will be auto-calculated from 'values' and 'colormap'.
        The rest of the explanation is in DFElement class.
        """
        super().__init__(*args, **kwargs)
        self.marker_dict = {
            'square': 0,
            'circle': 1,
            'sphere': 2,
        }

    def _fill_properties(self, *args, **kwargs) -> None:
        super()._fill_properties(*args, **kwargs)
        self.marker = kwargs.get('marker', 'square')
        self.point_size = kwargs.get('point_size', float(kwargs.get('avg_size', 1.0)))

    """
    Properties
    """
    @property
    def customizable_properties(self) -> list:
        return ['alpha', 'colormap', 'vmin', 'vmax', 'marker', 'avg_size']

    @property
    def exportable_properties(self) -> list:
        return ['alpha', 'colormap', 'headers', 'marker', 'avg_size']

    @property
    def avg_size(self) -> float:
        return self.point_size.mean()

    @property
    def point_size(self) -> np.ndarray:
        return self.datasets.get('size')

    @property
    def marker_num(self) -> int:
        return self.marker_dict.get(self.marker, 0)

    @property
    def marker(self) -> str:
        return self.properties.get('marker', 'square')

    @avg_size.setter
    def avg_size(self, _avg_size: float) -> None:
        self.point_size = _avg_size

    @point_size.setter
    def point_size(self, _size) -> None:
        if '__len__' in dir(_size):
            self.datasets['size'] = np.array(_size, np.float32)
        else:
            self.datasets['size'] = np.tile(_size, self.x.size).astype(np.float32)

    @marker.setter
    def marker(self, _marker: str) -> None:
        self.properties['marker'] = _marker



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtCore import Qt, QPoint
from qtpy.QtGui import QMouseEvent
from .mode import Mode


class NormalMode(Mode):
    def __init__(self):
        super().__init__()
        self.name = 'Normal Mode'
        self.lastPos = None

    def mousePressEvent(self, event: QMouseEvent, viewer) -> None:
        self.lastPos = QPoint(event.pos())

    def mouseDoubleClickEvent(self, event: QMouseEvent, viewer) -> None:
        viewer.detect_mesh_intersection(event.pos().x(), event.pos().y(), 1.0)

    def mouseMoveEvent(self, event: QMouseEvent, viewer) -> None:
        if self.lastPos is None:
            return

        dx = event.x() - self.lastPos.x()
        dy = event.y() - self.lastPos.y()
        smoothness = max(viewer.smoothness, 0.1)

        viewer.world.setToIdentity()
        if event.buttons() == Qt.LeftButton:
            self.set_x_rotation(viewer, viewer.xCenterRot + dy / smoothness)
            self.set_z_rotation(viewer, viewer.zCenterRot + dx / smoothness)
        elif event.buttons() == Qt.RightButton:
            self.set_x_rotation(viewer, viewer.xCenterRot + dy / smoothness)
            self.set_y_rotation(viewer, viewer.yCenterRot + dx / smoothness)
        elif event.buttons() == Qt.MiddleButton:
            off_center = viewer.off_center[2]
            distance_x = off_center / viewer.width()
            distance_y = off_center / viewer.height()
            self.set_x_movement(viewer, viewer.xCameraPos - (distance_x * dx))
            self.set_y_movement(viewer, viewer.yCameraPos + (distance_y * dy))

        self.lastPos = QPoint(event.pos())



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .shaderprogram import ShaderProgram


class PointProgram(ShaderProgram):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.base_name = 'Point'

    def setup(self) -> None:
        super().setup()
        self.add_uniform_loc('viewport')
        self.add_uniform_loc('marker')

    def inner_draw(self, drawables: list) -> None:
        for drawable in drawables:
            # We need DPI awareness for the size of the impostors
            viewport = [float(self.viewer.devicePixelRatio() * self.viewer.width()),
                        float(self.viewer.devicePixelRatio() * self.viewer.height())]
            self.update_uniform('viewport', *viewport)
            self.update_uniform('marker', drawable.element.marker_num)
            drawable.draw()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
from OpenGL.GL import *

from .gldrawable import GLDrawable


class MeshGL(GLDrawable):
    def __init__(self, element, *args, **kwargs):
        super().__init__(element, *args, **kwargs)
        self.indices_size = 0

        self._highlighted = kwargs.pop('highlight', False)
        self._wireframed = kwargs.pop('wireframe', False)

    """
    Properties
    """
    @property
    def is_highlighted(self) -> bool:
        return self._highlighted

    @property
    def is_wireframed(self) -> bool:
        return self._wireframed

    @property
    def is_turbo_ready(self) -> bool:
        return self.is_boostable and not self.is_highlighted and not self.is_wireframed

    @is_highlighted.setter
    def is_highlighted(self, status: bool) -> None:
        self._highlighted = status
        self.notify()

    @is_wireframed.setter
    def is_wireframed(self, status: bool) -> None:
        self._wireframed = status
        self.notify()

    """
    Quick MeshGL API
    """
    def enable_highlighting(self) -> None:
        self.is_highlighted = True

    def disable_highlighting(self) -> None:
        self.is_highlighted = False

    def toggle_highlighting(self) -> bool:
        self.is_highlighted = not self.is_highlighted
        return self.is_highlighted

    def enable_wireframe(self) -> None:
        self.is_wireframed = True

    def disable_wireframe(self) -> None:
        self.is_wireframed = False

    def toggle_wireframe(self) -> bool:
        self.is_wireframed = not self.is_wireframed
        return self.is_wireframed

    """
    Internal methods
    """
    def setup_attributes(self) -> None:
        _POSITION = 0
        _COLOR = 1

        # Generate VAO and VBOs (see GLDrawable)
        self.create_vao_vbos(3)

        # Data
        vertices = self.element.vertices.astype(np.float32)
        indices = self.element.indices.astype(np.uint32)
        colors = self.element.rgba.astype(np.float32)

        self.indices_size = indices.size

        glBindVertexArray(self.vao)

        # Fill buffers (see GLDrawable)
        self.fill_buffer(_POSITION, 3, vertices, GLfloat, GL_FLOAT, self.vbos[_POSITION])
        self.fill_buffer(_COLOR, 4, colors, GLfloat, GL_FLOAT, self.vbos[_COLOR])

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.vbos[-1])
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices, GL_STATIC_DRAW)

        # The attribute advances once per divisor instances of the set(s) of vertices being rendered
        # And guess what, we have just 1 instance, exactly what we wanted!
        glVertexAttribDivisor(_COLOR, 1)

        glBindVertexArray(0)

    def draw(self) -> None:
        glBindVertexArray(self.vao)
        glDrawElements(GL_TRIANGLES, self.indices_size, GL_UNSIGNED_INT, None)
        glBindVertexArray(0)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from OpenGL.GL import *

from ...model.elements.nullelement import NullElement


class GLDrawable:
    def __init__(self, element, *args, **kwargs):
        # assert element
        super().__setattr__('element', element or NullElement())  # self.element = element

        self.vaos = []
        self.vbos = []
        self.observers = []

        self._initialized = kwargs.pop('initialized', False)
        self._visible = kwargs.pop('visible', True)
        self._boostable = kwargs.pop('turbo', False)

    # Note: The following "hacks" are shortened versions of Delegator Pattern.
    # They're convenient, but optional.
    #
    # Example:
    # d = GLDrawable(element, *args, **kwargs)
    # assert d.alpha is d.element.alpha  => True
    def __dir__(self) -> list:
        # Hack to expose GLDrawable's attributes AND self.element's attributes
        # as if they were GLDrawable's attributes.
        # https://stackoverflow.com/q/15507848
        return sorted(set(dir(type(self)) + list(self.__dict__.keys()) + dir(self.element)))

    def __getattribute__(self, attr: str):
        # Hack to get our attributes.
        # If not found, search self.element's attributes.
        # https://stackoverflow.com/a/2405617
        try:
            return super().__getattribute__(attr)
        except AttributeError:
            return self.element.__getattribute__(attr)

    def __setattr__(self, key, value) -> None:
        # Hack to set our attributes.
        # We'll try to set our element's attribute first, then ourselves.
        # https://stackoverflow.com/a/7042247
        if key in dir(self.element):
            self.element.__setattr__(key, value)
        else:
            super().__setattr__(key, value)

    @property
    def vao(self) -> int:
        # We already know that we have only one VAO.
        # But cleanup is easier if we have the VAO in a list.
        return self.vaos[-1]

    def initialize(self) -> None:
        if not self.is_initialized:
            self.setup_attributes()
        self.is_initialized = True

    def setup_attributes(self) -> None:
        pass

    def create_vao_vbos(self, vbo_count: int) -> None:
        if len(self.vaos) == 0:
            self.vaos = [glGenVertexArrays(1)]
            self.vbos = glGenBuffers(vbo_count)

    @staticmethod
    def fill_buffer(pointer, basesize, array, glsize, gltype, vbo):
        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        glBufferData(GL_ARRAY_BUFFER, sizeof(glsize) * array.size, array, GL_STATIC_DRAW)
        glVertexAttribPointer(pointer, basesize, gltype, False, 0, None)
        glEnableVertexAttribArray(pointer)

    def draw(self) -> None:
        pass

    def cleanup(self) -> None:
        if self.is_initialized:
            glDeleteBuffers(len(self.vbos), self.vbos)
            glDeleteVertexArrays(len(self.vaos), self.vaos)

    """
    Properties
    """
    @property
    def is_initialized(self) -> bool:
        return self._initialized

    @property
    def is_visible(self) -> bool:
        return self._visible

    @property
    def is_boostable(self) -> bool:
        return self._boostable

    @is_initialized.setter
    def is_initialized(self, status: bool) -> None:
        self._initialized = status

    @is_visible.setter
    def is_visible(self, status: bool) -> None:
        self._visible = status
        self.notify()

    @is_boostable.setter
    def is_boostable(self, status: bool) -> None:
        self._boostable = status
        self.notify()

    """
    Quick GLDrawable API
    """
    def add_observer(self, observer) -> None:
        self.observers.append(observer)

    def notify(self) -> None:
        for observer in self.observers:
            observer.recreate()

    def show(self) -> None:
        self.is_visible = True

    def hide(self) -> None:
        self.is_visible = False

    def toggle_visibility(self) -> None:
        self.is_visible = not self.is_visible



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .glcollection import GLCollection

from ..drawables.axisgl import AxisGL
from ..glprograms.axisprogram import AxisProgram


class GLAxisCollection(GLCollection):
    def __init__(self, viewer=None):
        super().__init__(viewer)
        self.associate(AxisProgram(viewer), lambda: self.filter(AxisGL))



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np

from .gldrawable import GLDrawable
from OpenGL.GL import *


class AxisGL(GLDrawable):
    def __init__(self, element=None, *args, **kwargs):
        super().__init__(element, *args, **kwargs)

    def __dir__(self) -> list:
        return sorted(set(dir(type(self)) + list(self.__dict__.keys())))

    def setup_attributes(self) -> None:
        _POSITION = 0
        _COLOR = 1

        # Generate VAO and VBOs (see GLDrawable)
        self.create_vao_vbos(2)

        # Data
        vertices = np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0],
                             [0.0, 0.0, 0.0], [0.0, 1.0, 0.0],
                             [0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]).astype(np.float32)

        colors = np.array([[1.0, 0.0, 0.0], [1.0, 0.0, 0.0],
                           [0.0, 1.0, 0.0], [0.0, 1.0, 0.0],
                           [0.0, 0.0, 1.0], [0.0, 0.0, 1.0]]).astype(np.float32)

        glBindVertexArray(self.vao)

        # Fill buffers (see GLDrawable)
        self.fill_buffer(_POSITION, 3, vertices, GLfloat, GL_FLOAT, self.vbos[_POSITION])
        self.fill_buffer(_COLOR, 3, colors, GLfloat, GL_FLOAT, self.vbos[_COLOR])

        glBindVertexArray(0)

    def draw(self) -> None:
        glDisable(GL_DEPTH_TEST)
        glBindVertexArray(self.vao)
        glLineWidth(5)
        glDrawArrays(GL_LINES, 0, 6)
        glLineWidth(1)
        glBindVertexArray(0)
        glEnable(GL_DEPTH_TEST)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
from . import utils

from qtpy.QtCore import QDirIterator
from qtpy.QtCore import QFileInfo
from qtpy.QtCore import QMutex
from qtpy.QtCore import QMutexLocker

from .elements.element import Element
from .elements.elementcollection import ElementCollection
from .elements.blockelement import BlockElement
from .elements.pointelement import PointElement
from .elements.lineelement import LineElement
from .elements.meshelement import MeshElement
from .elements.tubeelement import TubeElement

from .parsers.parser import Parser
from .parsers.parsercollection import ParserCollection
from .parsers.dxfparser import DXFParser
from .parsers.offparser import OFFParser
from .parsers.h5mparser import H5MParser
from .parsers.h5pparser import H5PParser
from .parsers.csvparser import CSVParser
from .parsers.gslibparser import GSLibParser


class Model:
    def __init__(self):
        self._mutex = QMutex()
        self.parser_collection = ParserCollection()
        self.element_collection = ElementCollection()

        self.add_parser('dxf', DXFParser())
        self.add_parser('off', OFFParser())
        self.add_parser('h5m', H5MParser())
        self.add_parser('h5p', H5PParser())
        self.add_parser('csv', CSVParser())
        self.add_parser('out', GSLibParser())

    @property
    def last_id(self) -> int:
        return self.element_collection.last_id

    """
    Utilities
    """
    def add_parser(self, extension: str, handler: Parser) -> None:
        self.parser_collection.add(extension, handler)

    def get_parser(self, extension: str) -> Parser:
        return self.parser_collection.get(extension)

    @staticmethod
    def get_paths_from_directory(path: str) -> list:
        it = QDirIterator(path, QDirIterator.Subdirectories)
        path_list = []

        while it.hasNext():
            next_path = it.next()
            if QFileInfo(next_path).isFile():
                path_list.append(next_path)

        return sorted(path_list)

    """
    Generator methods
    """
    @staticmethod
    def generate_mesh(*args, **kwargs) -> MeshElement:
        return MeshElement(hint='mesh', *args, **kwargs)

    @staticmethod
    def generate_blocks(*args, **kwargs) -> BlockElement:
        return BlockElement(hint='block', *args, **kwargs)

    @staticmethod
    def generate_points(*args, **kwargs) -> PointElement:
        return PointElement(hint='point', *args, **kwargs)

    @staticmethod
    def generate_lines(*args, **kwargs) -> LineElement:
        return LineElement(hint='line', *args, **kwargs)

    @staticmethod
    def generate_tubes(*args, **kwargs) -> TubeElement:
        return TubeElement(hint='tube', *args, **kwargs)

    """
    Register methods
    """
    def register_element(self, element):
        # In a multi-threaded application, we can't risk assigning
        # the same ID to different elements, so we use a mutex here.
        with QMutexLocker(self._mutex):
            self.element_collection.add(element)

        return element

    def register_element_by_path(self, path: str, generator, *args, **kwargs):
        ext = path.split('.')[-1]
        info = self.get_parser(ext).load_file(path, *args, **kwargs)
        data = info.data
        properties = info.properties

        # Prioritize kwargs over file properties (useful in CLI)
        kwargs['data'] = kwargs.get('data', data)
        for k, v in properties.items():
            kwargs[k] = kwargs.get(k, v)

        return self.register_element(generator(*args, **kwargs))

    """
    Load methods by arguments
    """
    def mesh(self, *args, **kwargs) -> MeshElement:
        return self.register_element(self.generate_mesh(*args, **kwargs))

    def blocks(self, *args, **kwargs) -> BlockElement:
        return self.register_element(self.generate_blocks(*args, **kwargs))

    def points(self, *args, **kwargs) -> PointElement:
        return self.register_element(self.generate_points(*args, **kwargs))

    def lines(self, *args, **kwargs) -> LineElement:
        return self.register_element(self.generate_lines(*args, **kwargs))

    def tubes(self, *args, **kwargs) -> TubeElement:
        return self.register_element(self.generate_tubes(*args, **kwargs))

    """
    Load methods by path
    """
    def mesh_by_path(self, path: str, *args, **kwargs) -> MeshElement:
        return self.register_element_by_path(path, self.generate_mesh, *args, **kwargs)

    def blocks_by_path(self, path: str, *args, **kwargs) -> BlockElement:
        return self.register_element_by_path(path, self.generate_blocks, *args, **kwargs)

    def points_by_path(self, path: str, *args, **kwargs) -> PointElement:
        return self.register_element_by_path(path, self.generate_points, *args, **kwargs)

    def lines_by_path(self, path: str, *args, **kwargs) -> LineElement:
        return self.register_element_by_path(path, self.generate_lines, *args, **kwargs)

    def tubes_by_path(self, path: str, *args, **kwargs) -> TubeElement:
        return self.register_element_by_path(path, self.generate_tubes, *args, **kwargs)

    """
    Element handling
    """
    def get(self, _id: int) -> Element:
        return self.element_collection.get(_id)

    def delete(self, _id: int) -> None:
        self.element_collection.delete(_id)

    """
    Element exporting
    """
    def export(self, path: str, _id: int) -> None:
        element = self.get(_id)
        ext = path.split('.')[-1]

        data = element.data
        properties = {}
        for k in element.exportable_properties:
            properties[k] = getattr(element, k)

        self.get_parser(ext).save_file(path=path, data=data, properties=properties)

    def export_mesh(self, path: str, _id: int) -> None:
        self.export(path, _id)

    def export_blocks(self, path: str, _id: int) -> None:
        self.export(path, _id)

    def export_points(self, path: str, _id: int) -> None:
        self.export(path, _id)

    def export_lines(self, path: str, _id: int) -> None:
        self.export(path, _id)

    def export_tubes(self, path: str, _id: int) -> None:
        self.export(path, _id)

    """
    Adapter for viewer's utilities (slice/distance)
    """
    @staticmethod
    def slice_meshes(origin: np.ndarray, plane_normal: np.ndarray, meshes: list) -> dict:
        """
        Returns a dict with the following structure:

        {
            'plane_origin': list(float),
            'plane_normal': list(float)',
            'slices': [{
                'origin_id': int,
                'vertices': list(list(float))
            }]
        }
        """
        return {
            'plane_origin': origin,
            'plane_normal': plane_normal,
            'slices': [
                {
                    'origin_id': mesh.id,
                    'vertices': mesh.slice_with_plane(origin, plane_normal)
                } for mesh in meshes]
        }

    @staticmethod
    def slice_blocks(origin: np.ndarray, plane_normal: np.ndarray, block_list: list) -> dict:
        """
        Returns a dict with the following structure:

        {
            'plane_origin': list(float),
            'plane_normal': list(float)',
            'slices': [{
                'origin_id': int,
                'indices': list(int)
            }]
        }
        """
        return {
            'plane_origin': origin,
            'plane_normal': plane_normal,
            'slices': [
                {
                    'origin_id': block.id,
                    'indices': block.slice_with_plane(origin, plane_normal)
                } for block in block_list]
        }

    @staticmethod
    def measure_from_rays(origin_list: list, ray_list: list, meshes: list) -> dict:
        """
        Returns a dict with the following structure:

        {
            'point_a': list(float) or None,
            'point_b': list(float) or None',
            'distance': float or None
        }
        """
        points_A = []
        points_B = []

        closest_A = None
        closest_B = None

        # Detect intersections
        for mesh in meshes:
            int_A = mesh.intersect_with_ray(origin_list[0], ray_list[0])
            int_B = mesh.intersect_with_ray(origin_list[1], ray_list[1])

            # Discard non-intersections
            if int_A.size > 0:
                points_A.append(utils.closest_point_to(origin_list[0], int_A))

            if int_B.size > 0:
                points_B.append(utils.closest_point_to(origin_list[1], int_B))

        # Get closest points for each origin
        if len(points_A) > 0:
            points_A = np.vstack(points_A)
            closest_A = utils.closest_point_to(origin_list[0], points_A)

        if len(points_B) > 0:
            points_B = np.vstack(points_B)
            closest_B = utils.closest_point_to(origin_list[1], points_B)

        # Calculate distance if possible
        try:
            distance = np.linalg.norm(closest_B - closest_A)
        except TypeError:
            distance = None

        return {
            'point_a': closest_A,
            'point_b': closest_B,
            'distance': distance,
        }



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from OpenGL.GL import glViewport
from .shaderprogram import ShaderProgram


class AxisProgram(ShaderProgram):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.base_name = 'Axis'

    def draw(self) -> None:
        viewport = [self.viewer.devicePixelRatio() * self.viewer.width(),
                    self.viewer.devicePixelRatio() * self.viewer.height()]
        w, h = viewport

        glViewport(w // 100, h // 100, h // 6, h // 6)
        super().draw()
        glViewport(0, 0, viewport[0], viewport[1])



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import pathlib

from qtpy.QtGui import QOpenGLShader
from qtpy.QtGui import QOpenGLShaderProgram


class ShaderProgram:
    def __init__(self, viewer):
        self.base_name = None
        self.viewer = viewer
        self.shader_program = None
        self.shader_dir = f'{pathlib.Path(__file__).parent.parent}/drawables/shaders'
        self.uniform_locs = {}
        self.opaques = []
        self.transparents = []

    @property
    def drawables(self) -> list:
        return self.opaques + self.transparents

    def setup(self) -> None:
        if self.shader_program:
            return

        self.shader_program = QOpenGLShaderProgram(self.viewer.context())
        self.setup_shaders()

        self.add_uniform_loc('model_view_matrix')
        self.add_uniform_loc('proj_matrix')

    def setup_shaders(self) -> None:
        # Placeholders to avoid early garbage collection
        vs = self.enable_vertex_shader()
        fs = self.enable_fragment_shader()

        self.shader_program.link()

    def _enable_shader(self, shader_type, filename: str) -> QOpenGLShader:
        shader = QOpenGLShader(shader_type)
        shader.compileSourceFile(f'{self.shader_dir}/{self.base_name}/{filename}')
        self.shader_program.addShader(shader)
        return shader

    def enable_vertex_shader(self, filename='vertex.glsl') -> QOpenGLShader:
        return self._enable_shader(QOpenGLShader.Vertex, filename)

    def enable_fragment_shader(self, filename='fragment.glsl') -> QOpenGLShader:
        return self._enable_shader(QOpenGLShader.Fragment, filename)

    def enable_geometry_shader(self, filename='geometry.glsl') -> QOpenGLShader:
        return self._enable_shader(QOpenGLShader.Geometry, filename)

    def add_uniform_loc(self, loc_str) -> None:
        self.uniform_locs[loc_str] = self.shader_program.uniformLocation(loc_str)

    def update_uniform(self, loc_str, *values) -> None:
        self.shader_program.setUniformValue(self.uniform_locs[loc_str], *values)

    def set_drawables(self, drawables: list) -> None:
        self.opaques = [d for d in drawables if d.alpha >= 0.99]
        self.transparents = [d for d in drawables if d.alpha < 0.99]

        for drawable in self.drawables:
            drawable.initialize()

    def bind(self) -> None:
        self.shader_program.bind()

    def recreate(self) -> None:
        pass

    def inner_draw(self, drawables: list) -> None:
        for drawable in drawables:
            drawable.draw()

    def draw(self) -> None:
        self.inner_draw(self.opaques)

    def redraw(self) -> None:
        self.inner_draw(self.transparents)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import meshcut
import numpy as np

from .. import utils
from .element import Element


class MeshElement(Element):
    def __init__(self, *args, **kwargs):
        """
        MeshElement is a class inheriting from Element.

        {
            'data': {
                'x': list[float],
                'y': list[float],
                'z': list[float],
                'indices': list[int]
            }
            'properties': {
                'color': list[float],
                'alpha': float
            },
            'metadata': {
                'id': int,
                'name': str or None,
                'extension': str or None
            }
        }
        """
        super().__init__(*args, **kwargs)

    """
    Element filling
    """
    def _fill_element(self, *args, **kwargs) -> None:
        # Vertices
        msg = f'Data must contain ["x", "y", "z"] or "vertices", got {list(kwargs.keys())}.'

        if 'vertices' in kwargs.keys():
            self._fill_as_vertices(*args, **kwargs)
            self._fill_indices(*args, **kwargs)
        elif 'x' in kwargs.keys() and 'y' in kwargs.keys() and 'z' in kwargs.keys():
            self._fill_as_xyz(*args, **kwargs)
            self._fill_indices(*args, **kwargs)
        elif 'data' in kwargs.keys():
            self._fill_as_data(*args, **kwargs)
        else:
            raise KeyError(msg)

    def _fill_indices(self, *args, **kwargs) -> None:
        # Indices
        msg = f'Data must contain "indices", got {list(kwargs.keys())}.'

        if 'indices' in kwargs.keys() and 'data' not in kwargs.keys():
            self.indices = kwargs.get('indices', [])
        else:
            raise KeyError(msg)

    def _fill_as_data(self, *args, **kwargs) -> None:
        data = kwargs.get('data', {})
        self.vertices = data.get('vertices', [])
        self.indices = data.get('indices', [])

    def _check_integrity(self) -> None:
        super()._check_integrity()
        if self.x.size != self.indices.max() + 1:
            raise ValueError('Wrong number of indices for mesh.')

    """
    Data
    """
    @property
    def indices(self) -> np.array:
        return self.data.get('indices', np.empty(0))

    @indices.setter
    def indices(self, indices) -> None:
        # GL_UNSIGNED_INT = np.uint32
        self.data['indices'] = np.array(indices, np.uint32)

    """
    Utilities
    """
    @property
    def volume(self) -> float:
        # Idea from https://www.geometrictools.com/Documentation/PolyhedralMassProperties.pdf
        # Optimizations taken from https://github.com/mikedh/trimesh/blob/master/trimesh/triangles.py
        triangles = self.vertices[self.indices]

        vectors = np.diff(triangles, axis=1)
        crosses = np.cross(vectors[:, 0], vectors[:, 1])
        del vectors

        # This is equivalent but faster than triangles.sum(axis=1)
        f1 = triangles[:, 0, :] + triangles[:, 1, :] + triangles[:, 2, :]
        del triangles

        # Seems very similar to determinant calculation
        volume = (crosses[:, 0] * f1[:, 0]).sum() / 6.0
        del crosses
        del f1

        return abs(volume)

    def slice_with_plane(self, plane_origin: np.ndarray, plane_normal: np.ndarray) -> list:
        # This returns a list with the slices (in case we have a concave mesh)
        try:
            return meshcut.cross_section(self.vertices, self.indices, np.array(plane_origin), np.array(plane_normal))
        except AssertionError:
            # Meshcut doesn't want to slice
            print(f'WARNING: Mesh {self.name} (id = {self.id}) cannot be sliced, fix your mesh!')
            return []

    def intersect_with_ray(self, origin: np.ndarray, ray: np.ndarray) -> np.ndarray:
        # Early AABB detection test
        if not utils.aabb_intersection(origin, ray, *self.bounding_box):
            return np.empty(0)

        return self.vectorized_intersection(origin, ray)

    def vectorized_intersection(self, origin: np.ndarray, ray: np.ndarray) -> np.ndarray:
        # Idea taken from https://cadxfem.org/inf/Fast%20MinimumStorage%20RayTriangle%20Intersection.pdf
        # Code adapted from https://en.wikipedia.org/wiki/M%C3%B6ller%E2%80%93Trumbore_intersection_algorithm
        # Manually vectorized to benefit from numpy's performance.

        triangles = self.vertices[self.indices]

        # Get individual vertices of each triangle
        vertex0 = triangles[:, 0, :]
        vertex1 = triangles[:, 1, :]
        vertex2 = triangles[:, 2, :]

        edge1 = vertex1 - vertex0
        edge2 = vertex2 - vertex0

        # Note: Both np.inner(a, b).diagonal() and np.dot(a, b.T)
        # run out of memory, but (a * b).sum(axis=1) does not,
        # so we're using that one instead.
        h = np.cross(ray, edge2)
        a = (edge1 * h).sum(axis=1)

        mask = abs(a) > 1e-12  # False => Ray is parallel to triangle.

        # Result of division by zero used deliberately
        with np.errstate(divide='ignore', invalid='ignore'):
            f = 1.0 / a  # Can this be solved in a more elegant way?
            s = origin - vertex0
            u = f * (s * h).sum(axis=1)

            mask = (0.0 <= u) & (u <= 1.0) & mask

            q = np.cross(s, edge1)
            v = f * (ray * q).sum(axis=1)

            mask = (v >= 0.0) & (u + v <= 1.0) & mask

            # At this stage we can compute t to find out where the intersection point is on the line.
            t = f * (edge2 * q).sum(axis=1)

            mask = (t > 1e-12) & mask  # Ray intersections
        return origin + ray * t[mask].reshape(-1, 1)  # Here are the intersections.



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .shaderprogram import ShaderProgram


class TubeProgram(ShaderProgram):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.base_name = 'Tube'



#!/usr/bin/env python

from .csvparser import CSVParser
from .dxfparser import DXFParser
from .gslibparser import GSLibParser
from .h5mparser import H5MParser
from .h5pparser import H5PParser
from .offparser import OFFParser



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
import pandas as pd

from .element import Element
from ...model import utils


class DFElement(Element):
    __slots__ = ['_mapper', '_datasets']

    def __init__(self, *args, **kwargs):
        """
        DFElement (DataFrameElement) is a class designed with the following idea:
        (replace list for numpy arrays in the implementation)

        {
            'data': {
                'x': list[float],
                'y': list[float],
                'z': list[float],
                'values: list[float],
            },
            'datasets': {
                'color': list[list[float]] (Optional, auto-generated from 'values' if None)
            },
            'properties': {
                'headers': list[str],
                'vmin': float,
                'vmax': float,
                'alpha': float,
                'colormap': str (Optional, used from 'values')
            }
            'metadata': {
                'id': int,
                'name': str or None,
                'extension': str or None
            }
        }

        The 'data' dictionary has at least 4 keys, but it can be more than 4.
        It will be implemented as a Pandas DataFrame.
        They don't need to be named 'x, y, z, values' (see self.mapper).

        The 'datasets' dictionary was created because some properties
        might get too big to fit in an HDF5 attribute.
        It's expected to be used by children of this class.
        """
        # Base data
        self._data: pd.DataFrame = pd.DataFrame()
        self._datasets: dict = {}
        self._properties: dict = {}
        self._metadata: dict = {'id': -1}

        self._mapper: dict = {k: k for k in ['x', 'y', 'z', 'values']}
        super()._initialize(*args, **kwargs)

    """
    Element filling
    """
    def _fill_element(self, *args, **kwargs) -> None:
        # Base data
        msg = f'Data must contain ["x", "y", "z"], "vertices" or "data", got {list(kwargs.keys())}.'
        if 'data' in kwargs.keys():
            self._fill_as_data(*args, **kwargs)
        elif 'vertices' in kwargs.keys():
            self._fill_as_vertices(*args, **kwargs)
            self._fill_as_values(*args, **kwargs)
        elif 'x' in kwargs.keys() and 'y' in kwargs.keys() and 'z' in kwargs.keys():
            self._fill_as_xyz(*args, **kwargs)
            self._fill_as_values(*args, **kwargs)
        else:
            raise KeyError(msg)

    def _fill_as_values(self, *args, **kwargs) -> None:
        self.values = np.array(kwargs.get('values', np.empty(self.x.size)))

    def _fill_as_data(self, *args, **kwargs) -> None:
        self.data = kwargs.get('data')

    def _fill_properties(self, *args, **kwargs) -> None:
        self.headers = kwargs.get('headers', list(self.data.keys())[:4])
        self.alpha = kwargs.get('alpha', 1.0)
        self.colormap = kwargs.get('colormap', 'red-blue')  # red-blue (min is red, max is blue)
        self.color = kwargs.get('color', [])

        self.vmin = kwargs.get('vmin', self.values.min())
        self.vmax = kwargs.get('vmax', self.values.max())

    """
    Main accessors (Override)
    """
    @property
    def data(self) -> pd.DataFrame:
        return self._data

    @data.setter
    def data(self, _data: dict) -> None:
        self._data = pd.DataFrame(_data)

    @property
    def mapper(self) -> dict:
        return self._mapper

    @property
    def datasets(self) -> dict:
        return self._datasets

    """
    Data
    """
    @property
    def x(self) -> np.ndarray:
        return self.data[self.mapper.get('x')].to_numpy()

    @property
    def y(self) -> np.ndarray:
        return self.data[self.mapper.get('y')].to_numpy()

    @property
    def z(self) -> np.ndarray:
        return self.data[self.mapper.get('z')].to_numpy()

    @property
    def values(self) -> np.ndarray:
        return self.data[self.mapper.get('values')].to_numpy()

    @x.setter
    def x(self, _x: list) -> None:
        self.data[self.mapper.get('x')] = np.array(_x)

    @y.setter
    def y(self, _y: list) -> None:
        self.data[self.mapper.get('y')] = np.array(_y)

    @z.setter
    def z(self, _z: list) -> None:
        self.data[self.mapper.get('z')] = np.array(_z)

    @values.setter
    def values(self, _values) -> None:
        self.data[self.mapper.get('values')] = np.array(_values)

    """
    Properties
    """
    @property
    def color(self) -> np.ndarray:
        if self.datasets.get('color').size == 0:
            return utils.values_to_rgb(self.values, self.vmin, self.vmax, self.colormap)
        return self.datasets.get('color')

    @property
    def colormap(self) -> str:
        return self.properties.get('colormap')

    @property
    def vmin(self) -> float:
        return self.properties.get('vmin')

    @property
    def vmax(self) -> float:
        return self.properties.get('vmax')

    @color.setter
    def color(self, _colors: list) -> None:
        self.datasets['color'] = np.array(_colors)

    @property
    def all_headers(self) -> list:
        return list(self.data.keys())

    @property
    def headers(self) -> list:
        return list(self.mapper.values())

    @colormap.setter
    def colormap(self, _colormap: str) -> None:
        if utils.parse_colormap(_colormap):  # Empty list interpreted as False
            self.properties['colormap'] = _colormap

    @vmin.setter
    def vmin(self, _vmin: float) -> None:
        try:
            self.properties['vmin'] = float(_vmin)
        except Exception:
            self.properties['vmin'] = self.values.min()

    @vmax.setter
    def vmax(self, _vmax: float) -> None:
        try:
            self.properties['vmax'] = float(_vmax)
        except Exception:
            self.properties['vmax'] = self.values.max()

    @headers.setter
    def headers(self, _headers: list) -> None:
        self.mapper['x'], self.mapper['y'], self.mapper['z'], self.mapper['values'] = _headers



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import sys

from qtpy.QtCore import Qt
from qtpy.QtCore import QTimer
from qtpy.QtWidgets import QApplication

from .gui.iconcollection import IconCollection
from .integrableviewer import IntegrableViewer


class Viewer(IntegrableViewer):
    def __init__(self):
        self.app = QApplication(sys.argv)
        super().__init__()
        self.setWindowTitle('BlastSight (Viewer)')
        self.setWindowIcon(IconCollection.get('blastsight.png'))

    def show(self, detached: bool = False, timer: int = 0, autofit: bool = True) -> None:
        # This will auto-fit to screen when used from a script,
        # as it's reasonable to expect the figure to be shown
        # immediately, even if it's far from [0.0, 0.0, 0.0].
        self.set_autofit_status(autofit)
        super().show()

        if detached:
            # This allow us to detach the widget (convenient if running
            # the viewer in an interactive console).
            # WARNING: On PySide2, the viewer WILL freeze when the timer
            # runs out, but you can manually close it with `viewer.close()`.
            QTimer.singleShot(timer, self.app.quit)

        self.app.exec_()

    def take_screenshot(self, save_path=None, width=None, height=None) -> None:
        self.resize(width or self.width(), height or self.height())
        super().take_screenshot(save_path)

    def dragEnterEvent(self, event, *args, **kwargs) -> None:
        super().dragEnterEvent(event, *args, **kwargs)

    def dropEvent(self, event, *args, **kwargs) -> None:
        super().dropEvent(event, *args, **kwargs)
        self.camera_at(self.last_id)

    def keyPressEvent(self, event) -> None:
        def rotate(angle_list: list) -> None:
            self.rotation_angle += angle_list
            self.update()

        def move(pos_list: list) -> None:
            self.camera_position += pos_list
            self.update()

        shortcut_commands_dict = {
            Qt.Key_1: self.plan_view,
            Qt.Key_2: self.north_view,
            Qt.Key_3: self.east_view,
            Qt.Key_4: self.perspective_projection,
            Qt.Key_5: self.orthographic_projection,
            Qt.Key_Space: self.fit_to_screen,
            Qt.Key_Delete: lambda: self.delete(self.last_id),
            Qt.Key_W: lambda: rotate([-10.0, 0.0, 0.0]),
            Qt.Key_S: lambda: rotate([10.0, 0.0, 0.0]),
            Qt.Key_A: lambda: rotate([0.0, -10.0, 0.0]),
            Qt.Key_D: lambda: rotate([0.0, 10.0, 0.0]),
            Qt.Key_Q: lambda: rotate([0.0, 0.0, 10.0]),
            Qt.Key_E: lambda: rotate([0.0, 0.0, -10.0]),
            Qt.Key_Left: lambda: move([1.0, 0.0, 0.0]),
            Qt.Key_Right: lambda: move([-1.0, 0.0, 0.0]),
            Qt.Key_Up: lambda: move([0.0, -1.0, 0.0]),
            Qt.Key_Down: lambda: move([0.0, 1.0, 0.0]),
            Qt.Key_PageDown: lambda: move([0.0, 0.0, -1.0]),
            Qt.Key_PageUp: lambda: move([0.0, 0.0, 1.0]),
        }

        # Execute command based on event.key()
        shortcut_commands_dict.get(event.key(), lambda: None)()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtCore import Qt
from qtpy.QtGui import QMouseEvent
from .mode import Mode


class SliceMode(Mode):
    def __init__(self):
        super().__init__()
        self.name = 'Slice Mode'
        self.origins = []
        self.rays = []

    def mousePressEvent(self, event: QMouseEvent, viewer) -> None:
        if event.buttons() == Qt.LeftButton:
            self.detect_rays(event, viewer)
        else:
            self.origins.clear()
            self.rays.clear()

    def detect_rays(self, event: QMouseEvent, viewer) -> None:
        x, y, z = [event.pos().x(), event.pos().y(), 1.0]
        self.rays.append(viewer.ray_from_click(x, y, z))
        self.origins.append(viewer.origin_from_click(x, y, z))

        if len(self.rays) == 2:
            viewer.slice_from_rays(self.origins, self.rays)
            self.origins.clear()
            self.rays.clear()



#!/usr/bin/env python


#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtGui import QMouseEvent, QWheelEvent


class Mode:
    def __init__(self):
        self.name = 'Base mode'

    def mousePressEvent(self, event: QMouseEvent, viewer) -> None:
        pass

    def mouseMoveEvent(self, event: QMouseEvent, viewer) -> None:
        pass

    def mouseReleaseEvent(self, event: QMouseEvent, viewer) -> None:
        pass

    def mouseDoubleClickEvent(self, event: QMouseEvent, viewer) -> None:
        pass

    def wheelEvent(self, event: QWheelEvent, viewer) -> None:
        dy = event.angleDelta().y()
        sign = dy / abs(dy) if abs(dy) > 1e-12 else 0.0
        smoothness = max(viewer.smoothness, 0.1)

        # Arbitrary number, but dependent on viewer's smoothness
        rate = 0.4 / smoothness
        movement_rate = rate / (1.0 + max(0.0, min(sign, rate)))

        off_center = viewer.off_center[2]
        shift = -sign * movement_rate * max(-sign, off_center) / smoothness

        self.set_z_movement(viewer, viewer.zCameraPos + shift)

    @staticmethod
    def set_x_rotation(viewer, angle: float) -> None:
        viewer.xCenterRot = angle % 360

    @staticmethod
    def set_y_rotation(viewer, angle: float) -> None:
        viewer.yCenterRot = angle % 360

    @staticmethod
    def set_z_rotation(viewer, angle: float) -> None:
        viewer.zCenterRot = angle % 360

    @staticmethod
    def set_x_movement(viewer, position: float) -> None:
        viewer.xCameraPos = position

    @staticmethod
    def set_y_movement(viewer, position: float) -> None:
        viewer.yCameraPos = position

    @staticmethod
    def set_z_movement(viewer, position: float) -> None:
        viewer.zCameraPos = position



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .glcollection import GLCollection

from ..drawables.meshgl import MeshGL
from ..drawables.blockgl import BlockGL
from ..drawables.linegl import LineGL
from ..drawables.pointgl import PointGL
from ..drawables.tubegl import TubeGL

from ..glprograms.meshprogram import MeshProgram
from ..glprograms.wireprogram import WireProgram
from ..glprograms.blockprogram import BlockProgram
from ..glprograms.blocklegacyprogram import BlockLegacyProgram
from ..glprograms.lineprogram import LineProgram
from ..glprograms.pointprogram import PointProgram
from ..glprograms.tubeprogram import TubeProgram
from ..glprograms.turbomeshprogram import TurboMeshProgram


class GLDrawableCollection(GLCollection):
    def __init__(self, viewer=None):
        super().__init__(viewer)
        # Lines
        self.associate(LineProgram(viewer), lambda: self.filter(LineGL))

        # Tubes
        self.associate(TubeProgram(viewer), lambda: self.filter(TubeGL))

        # Points
        self.associate(PointProgram(viewer), lambda: self.filter(PointGL))

        # Blocks
        self.associate(BlockLegacyProgram(viewer), lambda: [
            x for x in self.filter(BlockGL) if x.is_legacy])

        self.associate(BlockProgram(viewer), lambda: [
            x for x in self.filter(BlockGL) if not x.is_legacy])

        # Meshes
        self.associate(TurboMeshProgram(viewer), lambda: [
            x for x in self.filter(MeshGL) if x.is_turbo_ready])
        self.associate(MeshProgram(viewer), lambda: [
            x for x in self.filter(MeshGL) if not x.is_turbo_ready and not x.is_wireframed])
        self.associate(WireProgram(viewer), lambda: [
            x for x in self.filter(MeshGL) if x.is_wireframed])



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .gldrawable import GLDrawable
from .axisgl import AxisGL
from .backgroundgl import BackgroundGL
from .blockgl import BlockGL
from .pointgl import PointGL
from .linegl import LineGL
from .meshgl import MeshGL
from .tubegl import TubeGL



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import pandas as pd
from qtpy.QtCore import QFileInfo
from .parserdata import ParserData
from .parser import Parser


class H5PParser(Parser):
    @staticmethod
    def load_file(path: str, *args, **kwargs) -> ParserData:
        store = pd.HDFStore(path, 'r')
        try:
            properties = store.get_storer('data').attrs.metadata
        except AttributeError:
            properties = {}

        # Metadata
        properties['name'] = QFileInfo(path).completeBaseName()
        properties['extension'] = QFileInfo(path).suffix()

        data = ParserData()
        data.data = store['data']
        data.properties = properties

        store.close()
        return data

    @staticmethod
    def save_file(*args, **kwargs) -> None:
        path = kwargs.get('path')

        if path is None:
            raise KeyError('Path missing.')

        data = kwargs.get('data', [])
        properties = kwargs.get('properties', {})

        store = pd.HDFStore(path, 'w')
        store.put('data', data)
        store.get_storer('data').attrs.metadata = properties
        store.close()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np

from .gldrawable import GLDrawable
from OpenGL.GL import *


class BlockGL(GLDrawable):
    def __init__(self, element, *args, **kwargs):
        super().__init__(element, *args, **kwargs)
        self.num_cubes = 0
        self._legacy = kwargs.pop('legacy', False)

    """
    Properties
    """
    @property
    def is_legacy(self) -> bool:
        return self._legacy

    @is_legacy.setter
    def is_legacy(self, status: bool) -> None:
        self._legacy = status
        self.is_initialized = False
        self.notify()

    def setup_attributes(self) -> None:
        _POSITION = 0
        _COLOR = 1
        _ALPHA = 2
        _TEMPLATE = 3

        # Generate VAO and VBOs (see GLDrawable)
        self.create_vao_vbos(4)

        # Data
        vertices = np.array(self.element.vertices, np.float32)
        colors = np.array(self.element.color, np.float32)
        alpha = np.array([self.element.alpha], np.float32)
        template = self.generate_cube(self.element.block_size)

        self.num_cubes = len(vertices)

        glBindVertexArray(self.vao)

        # Fill buffers (see GLDrawable)
        self.fill_buffer(_POSITION, 3, vertices, GLfloat, GL_FLOAT, self.vbos[_POSITION])
        self.fill_buffer(_COLOR, 3, colors, GLfloat, GL_FLOAT, self.vbos[_COLOR])
        self.fill_buffer(_ALPHA, 1, alpha, GLfloat, GL_FLOAT, self.vbos[_ALPHA])
        self.fill_buffer(_TEMPLATE, 3, template, GLfloat, GL_FLOAT, self.vbos[_TEMPLATE])

        # The attribute advances once per divisor instances of the set(s) of vertices being rendered.
        if self.is_legacy:
            glVertexAttribDivisor(_POSITION, 1)
            glVertexAttribDivisor(_COLOR, 1)
        else:
            glVertexAttribDivisor(_POSITION, 0)
            glVertexAttribDivisor(_COLOR, 0)

        glVertexAttribDivisor(_TEMPLATE, 0)
        glVertexAttribDivisor(_ALPHA, -1)

        glBindVertexArray(0)

    def draw(self) -> None:
        # Force legacy method if OpenGL < 3.3
        if not self.is_legacy and float(f'{glGetIntegerv(GL_MAJOR_VERSION)}.{glGetIntegerv(GL_MINOR_VERSION)}') < 3.3:
            self.is_legacy = True
            return

        glBindVertexArray(self.vao)

        if self.is_legacy:
            glDrawArraysInstanced(GL_TRIANGLES, 0, 36, self.num_cubes)
        else:
            glDrawArrays(GL_POINTS, 0, self.num_cubes)

        glBindVertexArray(0)

    @staticmethod
    def generate_cube(size: np.ndarray) -> np.ndarray:
        vertices = np.array([
            [-1, -1, -1],
            [+1, -1, -1],
            [+1, +1, -1],
            [-1, +1, -1],
            [-1, -1, +1],
            [+1, -1, +1],
            [+1, +1, +1],
            [-1, +1, +1],
        ]) * size * 0.5

        indices = np.array([
            [1, 2, 3],
            [1, 3, 0],
            [4, 5, 6],
            [4, 6, 7],
            [0, 1, 5],
            [0, 5, 4],
            [2, 3, 7],
            [2, 7, 6],
            [3, 0, 4],
            [3, 4, 7],
            [1, 2, 6],
            [1, 6, 5],
        ])

        return vertices[indices].astype(np.float32)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from collections import OrderedDict


class GLProgramCollection:
    def __init__(self):
        self._collection = OrderedDict()

    def associate(self, program_type, association) -> None:
        self._collection[program_type] = association

    def get_programs(self) -> dict.keys:
        return self._collection.keys()

    def get_associations(self) -> dict.values:
        return self._collection.values()

    def get_pairs(self) -> dict.items:
        return self._collection.items()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QVBoxLayout
from qtpy.QtWidgets import QWidget

from .toolbar import ToolBar
from .treewidget import TreeWidget
from .iconcollection import IconCollection
from ..integrableviewer import IntegrableViewer


class Container(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setAcceptDrops(True)

        self.toolbar = ToolBar(self)
        self.actions = self.toolbar.action_collection
        self.viewer = IntegrableViewer(self)

        self.treeWidget = TreeWidget(viewer=self.viewer)
        self.treeWidget.setWindowTitle('Drawables')

        self.verticalLayout = QVBoxLayout(self)
        self.verticalLayout.addWidget(self.viewer)
        self.verticalLayout.addWidget(self.toolbar)

        self.setWindowTitle('BlastSight (Container)')
        self.setWindowIcon(IconCollection.get('blastsight.png'))
        self.setMinimumSize(600, 500)

        self.toolbar.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.toolbar.insertAction(self.actions.action_plan_view, self.actions.action_camera_properties)
        self.toolbar.addAction(self.actions.action_quit)

        self.connect_actions()

    def connect_actions(self) -> None:
        self.toolbar.connect_tree(self.treeWidget)
        self.toolbar.connect_viewer(self.viewer)
        self.toolbar.connect_main_widget(self)
        self.viewer.signal_file_modified.connect(self.treeWidget.auto_refill)

    def dragEnterEvent(self, event, *args, **kwargs) -> None:
        self.viewer.dragEnterEvent(event, *args, **kwargs)

    def dropEvent(self, event, *args, **kwargs) -> None:
        self.viewer.dropEvent(event, *args, **kwargs)



#!/usr/bin/env python

import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='blastsight',
    version='0.4.1',
    author='Gabriel Sanhueza',
    description='A 3D visualization library oriented to mining applications',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/gsanhueza/BlastSight',
    packages=setuptools.find_packages(exclude=['tests*']),
    include_package_data=True,
    data_files=[
        ('share/applications', ['blastsight.desktop']),
        ('share/pixmaps', ['blastsight/view/gui/UI/icons/blastsight.png']),
    ],
    install_requires=[
        'numpy',
        'PyQt5',
        'dxfgrabber',
        'PyOpenGL',
        'pandas',
        'qtpy',
        'h5py',
        'tables',
        'meshcut',
        'colour',
    ],
    tests_require=[
        'pytest',
        'pytest-cov',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)




#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
from .dfelement import DFElement


class BlockElement(DFElement):
    def __init__(self, *args, **kwargs):
        """
        BlockElement is a class inheriting from PointElement (which inherits from DFElement).

        {
            'data': {
                'x': list[float],
                'y': list[float],
                'z': list[float],
                'values: list[float],
            },
            'datasets': {
                'color': list[list[float]] (Optional, auto-generated from 'values' if None)
            },
            'properties': {
                'headers': list[str],
                'vmin': float,
                'vmax': float,
                'alpha': float,
                'size': float,
                'colormap': str (Optional, used from 'values')
            }
            'metadata': {
                'id': int,
                'name': str or None,
                'extension': str or None
            }
        }

        In a BlockElement, all blocks have the same 3D size, but each block has its own color.
        If the user didn't specify colors, they will be auto-calculated from 'values' and 'colormap'.
        The rest of the explanation is in DFElement class.
        """
        super().__init__(*args, **kwargs)

    def _fill_properties(self, *args, **kwargs) -> None:
        super()._fill_properties(*args, **kwargs)
        if kwargs.get('autosize', True):
            self.block_size = kwargs.get('block_size', self.get_autosize())
        else:
            self.block_size = kwargs.get('block_size', [1.0, 1.0, 1.0])

    def get_autosize(self) -> np.ndarray:
        size = []

        for v in [self.x, self.y, self.z]:
            delta = np.abs(np.diff(v))
            size.append(float(min(delta[delta > 0.0], default=1.0)))

        return np.array(size)

    """
    Properties
    """
    @property
    def customizable_properties(self) -> list:
        return ['alpha', 'colormap', 'vmin', 'vmax', 'block_size']

    @property
    def exportable_properties(self) -> list:
        return ['alpha', 'colormap', 'headers', 'block_size']

    @property
    def block_size(self) -> np.ndarray:
        return self.properties.get('size')

    @block_size.setter
    def block_size(self, _size: list) -> None:
        self.properties['size'] = np.array(_size) if len(_size) == 3 else np.array(self.get_autosize())

    @property
    def bounding_box(self) -> tuple:
        lo, hi = super().bounding_box
        return lo - (self.block_size / 2), hi + (self.block_size / 2)

    """
    Utilities
    """
    def slice_with_plane(self,
                         plane_origin: np.ndarray,
                         plane_normal: np.ndarray) -> np.ndarray:
        """
        *** Plane Equation: ax + by + cz + d = 0 ***

        Where [a, b, c] = plane_normal
        Where [x, y, z] = plane_origin (or any point that we know belongs to the plane)

        With this, we can get `d`:
        dot([a, b, c], [x, y, z]) + d = 0
        d = -dot([a, b, c], [x, y, z])

        Since we have multiple vertices, it's easier to multiply plane_normal with
        each vertex, and manually sum them to get an array of dot products.

        But our points are blocks (they have 3D dimensions in `block_size`).
        That means we have to tolerate more points, so we create a threshold.

        *** Plane Inequation: abs(ax + by + cz + d) <= threshold ***

        We need to know how inclined is the plane normal to know our threshold.
        Let's say our block_size is [10, 10, 10] (half_block is [5, 5, 5]).

        If the plane touches one face of the cube, our threshold is [-5, +5] * np.sqrt(1.0).
        If the plane touches one edge of the cube, our threshold is [-5, +5] * np.sqrt(2.0).
        If the plane touches one vertex of the cube, our threshold is [-5, +5] * np.sqrt(3.0).

        Since a cube is symmetrical by axes, we don't really care about the plane normal's signs.
        Then, we'll calculate np.dot(abs(plane_normal), half_block) to know the maximum
        tolerable distance between the cube center and its projection on the plane.

        The projection idea comes from
        https://gdbooks.gitbooks.io/3dcollisions/content/Chapter2/static_aabb_plane.html
        """
        plane_normal /= np.linalg.norm(plane_normal)
        half_block = np.array(self.block_size) / 2
        vertices = self.vertices

        plane_d = -np.dot(plane_normal, plane_origin)
        threshold = np.dot(np.abs(plane_normal), half_block)

        # In this context, np.inner(a, b) returns the same as (a * b).sum(axis=1), but it's faster.
        # Luckily, we don't run out of memory like in vectorized_triangles_intersection.
        mask = np.abs(np.inner(plane_normal, vertices) + plane_d) <= threshold

        # If mask = [True, False, True], then mask.nonzero()[-1] = [0, 2]
        return mask.nonzero()[-1]



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .shaderprogram import ShaderProgram
from OpenGL.GL import *


class MeshProgram(ShaderProgram):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.base_name = 'Mesh'

    def draw(self) -> None:
        # Highlighted
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        glLineWidth(3)
        super().inner_draw(filter(lambda x: x.is_highlighted, self.opaques))
        glLineWidth(1)

        # Opaque
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        super().draw()

    def redraw(self) -> None:
        # Highlighted
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        glLineWidth(3)
        super().inner_draw(filter(lambda x: x.is_highlighted, self.transparents))
        glLineWidth(1)

        # Transparent
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glDepthMask(GL_FALSE)
        glEnable(GL_CULL_FACE)

        for gl_cull in [GL_FRONT, GL_BACK]:
            glCullFace(gl_cull)
            super().redraw()

        glDisable(GL_CULL_FACE)
        glDepthMask(GL_TRUE)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import pathlib

from qtpy.QtWidgets import QDialog
from .tools import uic


class AboutDialog(QDialog):
    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        uic.loadUi(f'{pathlib.Path(__file__).parent}/UI/aboutdialog.ui', self)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .element import Element


class NullElement(Element):
    def __init__(self, *args, **kwargs):
        """
        NullElement is a placeholder for AxisGL and BackgroundGL.
        It has no data associated.
        """
        super().__init__(*args, **kwargs)

    def _fill_element(self, *args, **kwargs) -> None:
        pass

    def _check_integrity(self) -> None:
        pass



#!/usr/bin/env python


#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

class ParserData:
    def __init__(self):
        """
        For meshes:
        {
            'data': {
                'vertices': list[list[float]],
                'indices': list[int]
            },
            'properties': {
                'color': list[float],
                'alpha': float
            }
        }

        For blocks/points:
        {
            'data': Pandas DataFrame,
            'properties': dict of properties
        }
        """
        self._data = {'data': {}, 'properties': {}}

    @property
    def vertices(self) -> list:
        return self._data.get('data', {}).get('vertices', [])

    @property
    def indices(self) -> list:
        return self._data.get('data', {}).get('indices', [])

    @property
    def data(self) -> dict:
        return self._data.get('data', {})

    @property
    def properties(self) -> dict:
        return self._data.get('properties', {})

    @vertices.setter
    def vertices(self, _data) -> None:
        self._data['data']['vertices'] = _data

    @indices.setter
    def indices(self, _data) -> None:
        self._data['data']['indices'] = _data

    @data.setter
    def data(self, _data) -> None:
        self._data['data'] = _data

    @properties.setter
    def properties(self, _data) -> None:
        self._data['properties'] = _data



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import pathlib

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog
from .tools import uic


class CameraDialog(QDialog):
    def __init__(self, viewer=None):
        QDialog.__init__(self, viewer)
        uic.loadUi(f'{pathlib.Path(__file__).parent}/UI/cameradialog.ui', self)

        # Avoids the QObject::startTimer warning (maybe)
        self.setAttribute(Qt.WA_DeleteOnClose)

        self.camera_position = viewer.camera_position
        self.rotation_angle = viewer.rotation_angle
        self.rotation_center = viewer.rotation_center

    @property
    def camera_position(self) -> list:
        return [self.doubleSpinBox_x.value(),
                self.doubleSpinBox_y.value(),
                self.doubleSpinBox_z.value()]

    @property
    def rotation_angle(self) -> list:
        return [self.doubleSpinBox_rot_x.value(),
                self.doubleSpinBox_rot_y.value(),
                self.doubleSpinBox_rot_z.value()]

    @property
    def rotation_center(self) -> list:
        return [self.doubleSpinBox_center_x.value(),
                self.doubleSpinBox_center_y.value(),
                self.doubleSpinBox_center_z.value()]

    @camera_position.setter
    def camera_position(self, position: list) -> None:
        self.doubleSpinBox_x.setValue(position[0])
        self.doubleSpinBox_y.setValue(position[1])
        self.doubleSpinBox_z.setValue(position[2])

    @rotation_angle.setter
    def rotation_angle(self, angle: list) -> None:
        self.doubleSpinBox_rot_x.setValue(angle[0])
        self.doubleSpinBox_rot_y.setValue(angle[1])
        self.doubleSpinBox_rot_z.setValue(angle[2])

    @rotation_center.setter
    def rotation_center(self, center: list) -> None:
        self.doubleSpinBox_center_x.setValue(center[0])
        self.doubleSpinBox_center_y.setValue(center[1])
        self.doubleSpinBox_center_z.setValue(center[2])
